from argparse import ArgumentParser
from typing import Optional

from textual.app import App

import mys_goods_tool.user_data
from mys_goods_tool.exchange_mode import exchange_mode_simple
from mys_goods_tool.user_data import load_config

USAGE = """
Mys_Goods_Tool
使用说明：
%(prog)s [-m <运行模式>] [-c <用户数据文件路径>]
选项：
    -h, --help 显示此帮助信息
    -m, --mode <参数> 指定运行模式
        guide TUI指引模式，包含登陆绑定、管理兑换计划和开始兑换等功能（默认）
        exchange-simple 兑换模式，无TUI界面，仅输出日志文本
    -c, --conf <参数> 指定用户数据文件路径
例如：
    %(prog)s -m exchange-simple -c ./workplace/user_data.json
        通过该命令运行本程序，将读取 ./workplace/user_data.json 用户数据文件，并直接进入无TUI界面的兑换模式，等待到达兑换时间并执行兑换。
    %(prog)s
        通过该命令运行本程序或直接双击打开程序，将读取程序目录下的用户数据文件user_data.json，并提供登录绑定、管理兑换计划等功能。
        """.strip()


class ArgumentParserWithHelp(ArgumentParser):
    def error(self, message):
        self.print_help()
        super().error(message)


arg_parser = ArgumentParserWithHelp(description="Mys_Goods_Tool", usage=USAGE)
arg_parser.add_argument("-m", "--mode", dest="mode", choices=["guide", "exchange-simple"],
                        default="guide")
arg_parser.add_argument("-c", "--conf", type=str, dest="conf", default=None)

TEXTUAL_DEBUG = False


def main(textual_app: Optional[App] = None):
    arg = arg_parser.parse_args()

    if arg.conf is not None:
        mys_goods_tool.user_data.CONFIG_PATH = arg.conf
        mys_goods_tool.user_data.config, mys_goods_tool.user_data.different_device_and_salt = load_config()
    if arg.mode == "guide":
        if textual_app is None:
            from mys_goods_tool.tui import TuiApp
            textual_app = TuiApp()
        textual_app.run()
    elif arg.mode == "exchange-simple":
        exchange_mode_simple()


if __name__ == "__main__":
    app: Optional[App] = None
    if TEXTUAL_DEBUG:
        from mys_goods_tool.tui import TuiApp

        app = TuiApp()
    main(app)



import time
from typing import List, Optional, Tuple, Dict, Any, Union, Type
from urllib.parse import urlencode

import httpx
import tenacity
from pydantic import ValidationError, BaseModel
from requests.utils import dict_from_cookiejar

from mys_goods_tool.data_model import GameRecord, GameInfo, Good, Address, BaseApiStatus, MmtData, GeetestResult, \
    GetCookieStatus, \
    CreateMobileCaptchaStatus, GetGoodDetailStatus, ExchangeStatus, GeetestResultV4, GetFpStatus
from mys_goods_tool.user_data import config as conf, UserAccount, BBSCookies, ExchangePlan, ExchangeResult
from mys_goods_tool.utils import generate_device_id, logger, generate_ds, Subscribe, \
    NtpTime, get_async_retry, generate_seed_id, generate_fp_locally

URL_LOGIN_TICKET_BY_CAPTCHA = "https://webapi.account.mihoyo.com/Api/login_by_mobilecaptcha"
URL_LOGIN_TICKET_BY_PASSWORD = "https://webapi.account.mihoyo.com/Api/login_by_password"
URL_MULTI_TOKEN_BY_LOGIN_TICKET = "https://api-takumi.mihoyo.com/auth/api/getMultiTokenByLoginTicket?login_ticket={0}&token_types=3&uid={1}"
URL_COOKIE_TOKEN_BY_CAPTCHA = "https://api-takumi.mihoyo.com/account/auth/api/webLoginByMobile"
URL_COOKIE_TOKEN_BY_STOKEN = "https://passport-api.mihoyo.com/account/auth/api/getCookieAccountInfoBySToken"
URL_LTOKEN_BY_STOKEN = "https://passport-api.mihoyo.com/account/auth/api/getLTokenBySToken"
URL_STOKEN_V2_BY_V1 = "https://passport-api.mihoyo.com/account/ma-cn-session/app/getTokenBySToken"
URL_ACTION_TICKET = "https://api-takumi.mihoyo.com/auth/api/getActionTicketBySToken?action_type=game_role&stoken={stoken}&uid={bbs_uid}"
URL_GAME_RECORD = "https://api-takumi-record.mihoyo.com/game_record/card/wapi/getGameRecordCard?uid={}"
URL_GAME_LIST = "https://bbs-api.mihoyo.com/apihub/api/getGameList"
URL_MYB = "https://api-takumi.mihoyo.com/common/homutreasure/v1/web/user/point?app_id=1&point_sn=myb"
URL_DEVICE_LOGIN = "https://bbs-api.mihoyo.com/apihub/api/deviceLogin"
URL_DEVICE_SAVE = "https://bbs-api.mihoyo.com/apihub/api/saveDevice"
URL_GOOD_LIST = "https://api-takumi.mihoyo.com/mall/v1/web/goods/list?app_id=1&point_sn=myb&page_size=20&page={" \
                "page}&game={game} "
URL_CHECK_GOOD = "https://api-takumi.mihoyo.com/mall/v1/web/goods/detail?app_id=1&point_sn=myb&goods_id={}"
URL_EXCHANGE = "https://api-takumi.miyoushe.com/mall/v1/web/goods/exchange"
URL_ADDRESS = "https://api-takumi.mihoyo.com/account/address/list?t={}"
URL_REGISTRABLE = "https://webapi.account.mihoyo.com/Api/is_mobile_registrable?mobile={mobile}&t={t}"
URL_CREATE_MMT = "https://webapi.account.mihoyo.com/Api/create_mmt?scene_type=1&now={now}&reason=user.mihoyo.com%2523%252Flogin%252Fcaptcha&action_type=login_by_mobile_captcha&t={t}"
URL_CREATE_MOBILE_CAPTCHA = "https://webapi.account.mihoyo.com/Api/create_mobile_captcha"
URL_GET_USER_INFO = "https://bbs-api.miyoushe.com/user/api/getUserFullInfo?uid={uid}"
URL_GET_DEVICE_FP = "https://public-data-api.mihoyo.com/device-fp/api/getFp"

HEADERS_WEBAPI = {
    "Host": "webapi.account.mihoyo.com",
    "Connection": "keep-alive",
    "sec-ch-ua": conf.device_config.UA,
    "DNT": "1",
    "x-rpc-device_model": conf.device_config.X_RPC_DEVICE_MODEL_PC,
    "sec-ch-ua-mobile": "?0",
    "User-Agent": conf.device_config.USER_AGENT_PC,
    "x-rpc-device_id": None,
    "Accept": "application/json, text/plain, */*",
    "x-rpc-device_name": conf.device_config.X_RPC_DEVICE_NAME_PC,
    "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
    "x-rpc-client_type": "4",
    "sec-ch-ua-platform": conf.device_config.UA_PLATFORM,
    "Origin": "https://user.mihoyo.com",
    "Sec-Fetch-Site": "same-site",
    "Sec-Fetch-Mode": "cors",
    "Sec-Fetch-Dest": "empty",
    "Referer": "https://user.mihoyo.com/",
    "Accept-Encoding": "gzip, deflate, br",
    "Accept-Language": "zh-CN,zh;q=0.9,en;q=0.8,en-GB;q=0.7,en-US;q=0.6"
}
HEADERS_PASSPORT_API = {
    "Host": "passport-api.mihoyo.com",
    "Content-Type": "application/json",
    "Accept": "*/*",
    # "x-rpc-device_fp": "",
    "x-rpc-client_type": "1",
    "x-rpc-device_id": None,
    # "x-rpc-app_id": "bll8iq97cem8",
    "Accept-Language": "zh-CN,zh-Hans;q=0.9",
    "x-rpc-game_biz": "bbs_cn",
    "Accept-Encoding": "gzip, deflate, br",
    "x-rpc-device_model": conf.device_config.X_RPC_DEVICE_MODEL_MOBILE,
    "User-Agent": conf.device_config.USER_AGENT_OTHER,
    "x-rpc-device_name": conf.device_config.X_RPC_DEVICE_NAME_MOBILE,
    "x-rpc-app_version": conf.device_config.X_RPC_APP_VERSION,
    # 抓包时 "2.47.1"

    "x-rpc-sdk_version": "1.6.1",
    "Connection": "keep-alive",
    "x-rpc-sys_version": conf.device_config.X_RPC_SYS_VERSION
}
HEADERS_API_TAKUMI_PC = {
    "Host": "api-takumi.mihoyo.com",
    "Content-Type": "application/json;charset=utf-8",
    "Origin": "https://bbs.mihoyo.com",
    "Accept-Encoding": "gzip, deflate, br",
    "Connection": "keep-alive",
    "Accept": "application/json, text/plain, */*",
    "User-Agent": conf.device_config.USER_AGENT_PC,
    "Referer": "https://bbs.mihoyo.com/",
    "Accept-Language": "zh-CN,zh-Hans;q=0.9"
}
HEADERS_ACTION_TICKET = {
    "Host": "api-takumi.mihoyo.com",
    "x-rpc-device_model": conf.device_config.X_RPC_DEVICE_MODEL_MOBILE,
    "User-Agent": conf.device_config.USER_AGENT_OTHER,
    "Referer": "https://webstatic.mihoyo.com/",
    "x-rpc-device_name": conf.device_config.X_RPC_DEVICE_NAME_MOBILE,
    "Origin": "https://webstatic.mihoyo.com",
    "Content-Length": "66",
    "Connection": "keep-alive",
    "x-rpc-channel": conf.device_config.X_RPC_CHANNEL,
    "x-rpc-app_version": conf.device_config.X_RPC_APP_VERSION,
    "Accept-Language": "zh-CN,zh-Hans;q=0.9",
    "DS": None,
    "x-rpc-device_id": None,
    "x-rpc-client_type": "5",
    "Accept": "application/json, text/plain, */*",
    "Content-Type": "application/json;charset=utf-8",
    "Accept-Encoding": "gzip, deflate, br",
    "x-rpc-sys_version": conf.device_config.X_RPC_SYS_VERSION,
    "x-rpc-platform": conf.device_config.X_RPC_PLATFORM
}
HEADERS_GAME_RECORD = {
    "Host": "api-takumi-record.mihoyo.com",
    "Origin": "https://webstatic.mihoyo.com",
    "Connection": "keep-alive",
    "Accept": "application/json, text/plain, */*",
    "User-Agent": conf.device_config.USER_AGENT_MOBILE,
    "Accept-Language": "zh-CN,zh-Hans;q=0.9",
    "Referer": "https://webstatic.mihoyo.com/",
    "Accept-Encoding": "gzip, deflate, br"
}
HEADERS_GAME_LIST = {
    "Host": "bbs-api.mihoyo.com",
    "DS": None,
    "Accept": "*/*",
    "x-rpc-device_id": generate_device_id(),
    "x-rpc-client_type": "1",
    "x-rpc-channel": conf.device_config.X_RPC_CHANNEL,
    "Accept-Language": "zh-CN,zh-Hans;q=0.9",
    "Accept-Encoding": "gzip, deflate, br",
    "x-rpc-sys_version": conf.device_config.X_RPC_SYS_VERSION,
    "Referer": "https://app.mihoyo.com",
    "x-rpc-device_name": conf.device_config.X_RPC_DEVICE_NAME_MOBILE,
    "x-rpc-app_version": conf.device_config.X_RPC_APP_VERSION,
    "User-Agent": conf.device_config.USER_AGENT_OTHER,
    "Connection": "keep-alive",
    "x-rpc-device_model": conf.device_config.X_RPC_DEVICE_MODEL_MOBILE
}
HEADERS_MYB = {
    "Host": "api-takumi.mihoyo.com",
    "Origin": "https://webstatic.mihoyo.com",
    "Connection": "keep-alive",
    "Accept": "application/json, text/plain, */*",
    "User-Agent": conf.device_config.USER_AGENT_MOBILE,
    "Accept-Language": "zh-CN,zh-Hans;q=0.9",
    "Referer": "https://webstatic.mihoyo.com/",
    "Accept-Encoding": "gzip, deflate, br"
}
HEADERS_DEVICE = {
    "DS": None,
    "x-rpc-client_type": "2",
    "x-rpc-app_version": conf.device_config.X_RPC_APP_VERSION,
    "x-rpc-sys_version": conf.device_config.X_RPC_SYS_VERSION_ANDROID,
    "x-rpc-channel": conf.device_config.X_RPC_CHANNEL_ANDROID,
    "x-rpc-device_id": None,
    "x-rpc-device_name": conf.device_config.X_RPC_DEVICE_NAME_ANDROID,
    "x-rpc-device_model": conf.device_config.X_RPC_DEVICE_MODEL_ANDROID,
    "Referer": "https://app.mihoyo.com",
    "Content-Type": "application/json; charset=UTF-8",
    "Host": "bbs-api.mihoyo.com",
    "Connection": "Keep-Alive",
    "Accept-Encoding": "gzip",
    "User-Agent": conf.device_config.USER_AGENT_ANDROID_OTHER
}
HEADERS_GOOD_LIST = {
    "Host":
        "api-takumi.mihoyo.com",
    "Accept":
        "application/json, text/plain, */*",
    "Origin":
        "https://user.mihoyo.com",
    "Connection":
        "keep-alive",
    "x-rpc-device_id": generate_device_id(),
    "x-rpc-client_type":
        "5",
    "User-Agent":
        conf.device_config.USER_AGENT_MOBILE,
    "Referer":
        "https://user.mihoyo.com/",
    "Accept-Language":
        "zh-CN,zh-Hans;q=0.9",
    "Accept-Encoding":
        "gzip, deflate, br"
}
HEADERS_EXCHANGE = {
    "Accept":
        "application/json, text/plain, */*",
    "Accept-Encoding":
        "gzip, deflate, br",
    "Accept-Language":
        "zh-CN,zh-Hans;q=0.9",
    "Connection":
        "keep-alive",
    "Content-Type":
        "application/json",
    "Host":
        "api-takumi.miyoushe.com",
    "Origin":
        "https://webstatic.miyoushe.com",
    "Referer":
        "https://webstatic.miyoushe.com/",
    "User-Agent":
        conf.device_config.USER_AGENT_MOBILE,
    "x-rpc-app_version":
        conf.device_config.X_RPC_APP_VERSION,
    "x-rpc-channel":
        "appstore",
    "x-rpc-client_type":
        "1",
    "x-rpc-verify_key":
        "bll8iq97cem8",
    "x-rpc-device_fp": None,
    "x-rpc-device_id": None,
    "x-rpc-device_model":
        conf.device_config.X_RPC_DEVICE_MODEL_MOBILE,
    "x-rpc-device_name":
        conf.device_config.X_RPC_DEVICE_NAME_MOBILE,
    "x-rpc-sys_version":
        conf.device_config.X_RPC_SYS_VERSION
}
HEADERS_ADDRESS = {
    "Host": "api-takumi.mihoyo.com",
    "Accept": "application/json, text/plain, */*",
    "Origin": "https://user.mihoyo.com",
    "Connection": "keep-alive",
    "x-rpc-device_id": None,
    "x-rpc-client_type": "5",
    "User-Agent": conf.device_config.USER_AGENT_MOBILE,
    "Referer": "https://user.mihoyo.com/",
    "Accept-Language": "zh-CN,zh-Hans;q=0.9",
    "Accept-Encoding": "gzip, deflate, br"
}

IncorrectReturn = (KeyError, TypeError, AttributeError, IndexError, ValidationError)
"""米游社API返回数据无效会触发的异常组合"""


def is_incorrect_return(exception: Exception, *addition_exceptions: Type[Exception]) -> bool:
    """
    判断是否是米游社API返回数据无效的异常
    :param exception: 异常对象
    :param addition_exceptions: 额外的异常类型，用于触发判断
    """
    """
        return exception in IncorrectReturn or
            exception.__cause__ in IncorrectReturn or
            isinstance(exception, IncorrectReturn) or
            isinstance(exception.__cause__, IncorrectReturn)
    """
    exceptions = IncorrectReturn + addition_exceptions
    return isinstance(exception, exceptions) or isinstance(exception.__cause__, exceptions)


class ApiResultHandler(BaseModel):
    """
    API返回的数据处理器
    """
    content: Dict[str, Any]
    """API返回的JSON对象序列化以后的Dict对象"""
    data: Optional[Dict[str, Any]]
    """API返回的数据体"""
    message: Optional[str]
    """API返回的消息内容"""
    retcode: Optional[int]
    """API返回的状态码"""

    def __init__(self, content: Dict[str, Any]):
        super().__init__(content=content)

        self.data = self.content.get("data")

        for key in ["retcode", "status"]:
            if not self.retcode:
                self.retcode = self.content.get(key)
                if not self.retcode:
                    self.retcode = self.data.get(key) if self.data else None

        self.message: Optional[str] = None
        for key in ["message", "msg"]:
            if not self.message:
                self.message = self.content.get(key)
                if not self.message:
                    self.message = self.data.get(key) if self.data else None

    @property
    def success(self):
        """
        是否成功
        """
        return self.retcode == 1 or self.message in ["成功", "OK"]

    @property
    def wrong_captcha(self):
        """
        是否返回验证码错误
        """
        return self.retcode in [-201, -302] or self.message in ["验证码错误", "Captcha not match Err"]

    @property
    def login_expired(self):
        """
        是否返回登录失效
        """
        return self.retcode in [-100, 10001] or self.message in ["登录失效，请重新登录"]

    @property
    def invalid_ds(self):
        """
        Headers里的DS是否无效
        """
        # TODO 2023/4/13: 待补充状态码
        #  return True if self.retcode == -... or self.message in ["invalid request"] else False
        return self.message in ["invalid request"]


async def get_game_record(account: UserAccount, retry: bool = True) -> Tuple[BaseApiStatus, Optional[List[GameRecord]]]:
    """
    获取用户绑定的游戏账户信息，返回一个GameRecord对象的列表

    :param account: 用户账户数据
    :param retry: 是否允许重试
    """
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    res = await client.get(URL_GAME_RECORD.format(account.bbs_uid), headers=HEADERS_GAME_RECORD,
                                           cookies=account.cookies.dict(), timeout=conf.preference.timeout)
                api_result = ApiResultHandler(res.json())
                if api_result.login_expired:
                    logger.info(
                        f"获取用户游戏数据(GameRecord) - 用户 {account.bbs_uid} 登录失效")
                    logger.debug(f"网络请求返回: {res.text}")
                    return BaseApiStatus(login_expired=True), None
                return BaseApiStatus(success=True), list(
                    map(GameRecord.parse_obj, api_result.data["list"]))
    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"获取用户游戏数据(GameRecord) - 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return BaseApiStatus(incorrect_return=True), None
        else:
            logger.exception(f"获取用户游戏数据(GameRecord) - 请求失败")
            return BaseApiStatus(network_error=True), None


async def get_game_list(retry: bool = True) -> Tuple[BaseApiStatus, Optional[List[GameInfo]]]:
    """
    获取米哈游游戏的详细信息，若返回`None`说明获取失败

    :param retry: 是否允许重试
    """
    headers = HEADERS_GAME_LIST.copy()
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                headers["DS"] = generate_ds()
                async with httpx.AsyncClient() as client:
                    res = await client.get(URL_GAME_LIST, headers=headers, timeout=conf.preference.timeout)
                api_result = ApiResultHandler(res.json())
                return BaseApiStatus(success=True), list(
                    map(GameInfo.parse_obj, api_result.data["list"]))
    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"获取游戏信息(GameInfo) - 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return BaseApiStatus(incorrect_return=True), None
        else:
            logger.exception(f"获取游戏信息(GameInfo) - 请求失败")
            return BaseApiStatus(network_error=True), None


async def get_user_myb(account: UserAccount, retry: bool = True) -> Tuple[BaseApiStatus, Optional[int]]:
    """
    获取用户当前米游币数量

    :param account: 用户账户数据
    :param retry: 是否允许重试
    """
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    res = await client.get(URL_MYB, headers=HEADERS_MYB,
                                           cookies=account.cookies.dict(v2_stoken=True, cookie_type=True),
                                           timeout=conf.preference.timeout)
                api_result = ApiResultHandler(res.json())
                if api_result.login_expired:
                    logger.info(
                        f"获取用户米游币 - 用户 {account.bbs_uid} 登录失效")
                    logger.debug(f"网络请求返回: {res.text}")
                    return BaseApiStatus(login_expired=True), None
                return BaseApiStatus(success=True), int(api_result.data["points"])
    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"获取用户米游币 - 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return BaseApiStatus(incorrect_return=True), None
        else:
            logger.exception(f"获取用户米游币 - 请求失败")
            return BaseApiStatus(network_error=True), None


async def device_login(account: UserAccount, retry: bool = True):
    """
    设备登录(deviceLogin)(适用于安卓设备)

    :param account: 用户账户数据
    :param retry: 是否允许重试
    """
    data = {
        "app_version": conf.device_config.X_RPC_APP_VERSION,
        "device_id": account.device_id_android,
        "device_name": conf.device_config.X_RPC_DEVICE_NAME_ANDROID,
        "os_version": "30",
        "platform": "Android",
        "registration_id": "1a0018970a5c00e814d"
    }
    headers = HEADERS_DEVICE.copy()
    headers["x-rpc-device_id"] = account.device_id_android
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                headers["DS"] = generate_ds(data)
                async with httpx.AsyncClient() as client:
                    res = await client.post(URL_DEVICE_LOGIN, headers=headers, json=data,
                                            cookies=account.cookies.dict(v2_stoken=True, cookie_type=True),
                                            timeout=conf.preference.timeout)
                api_result = ApiResultHandler(res.json())
                if api_result.login_expired:
                    logger.info(
                        f"设备登录(device_login) - 用户 {account.bbs_uid} 登录失效")
                    logger.debug(f"网络请求返回: {res.text}")
                    return BaseApiStatus(login_expired=True)
                if res.json()["message"] != "OK":
                    raise ValueError
                else:
                    return BaseApiStatus(success=True)
    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"设备登录(device_login) - 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return BaseApiStatus(incorrect_return=True)
        else:
            logger.exception(f"设备登录(device_login) - 请求失败")
            return BaseApiStatus(network_error=True)


async def device_save(account: UserAccount, retry: bool = True):
    """
    设备保存(saveDevice)(适用于安卓设备)

    :param account: 用户账户数据
    :param retry: 是否允许重试
    """
    data = {
        "app_version": conf.device_config.X_RPC_APP_VERSION,
        "device_id": account.device_id_android,
        "device_name": conf.device_config.X_RPC_DEVICE_NAME_ANDROID,
        "os_version": "30",
        "platform": "Android",
        "registration_id": "1a0018970a5c00e814d"
    }
    headers = HEADERS_DEVICE.copy()
    headers["x-rpc-device_id"] = account.device_id_android
    try:
        subscribe = Subscribe()
        async for attempt in get_async_retry(retry):
            with attempt:
                headers["DS"] = generate_ds(data)
                async with httpx.AsyncClient() as client:
                    res = await client.post(
                        URL_DEVICE_SAVE,
                        headers=headers,
                        json=data,
                        cookies=account.cookies.dict(v2_stoken=True, cookie_type=True),
                        timeout=conf.preference.timeout
                    )
                api_result = ApiResultHandler(res.json())
                if api_result.login_expired:
                    logger.info(
                        f"设备保存(device_save) - 用户 {account.bbs_uid} 登录失效")
                    logger.debug(f"网络请求返回: {res.text}")
                    return BaseApiStatus(login_expired=True)
                if res.json()["message"] != "OK":
                    raise ValueError
                else:
                    return BaseApiStatus(success=True)
    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"设备保存(device_save) - 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return BaseApiStatus(incorrect_return=True)
        else:
            logger.exception(f"设备保存(device_save) - 请求失败")
            return BaseApiStatus(network_error=True)


async def get_good_detail(good: Union[Good, str], retry: bool = True) -> Tuple[GetGoodDetailStatus, Optional[Good]]:
    """
    获取某商品的详细信息

    :param good: 商品对象 / 商品ID，如果指定为商品对象，则会更新商品对象的数据并返回其引用
    :param retry: 是否允许重试
    :return: 商品数据
    """
    good_id = good.goods_id if isinstance(good, Good) else good
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    res = await client.get(URL_CHECK_GOOD.format(good_id), timeout=conf.preference.timeout)
                api_result = ApiResultHandler(res.json())
                # TODO 2023/4/13: 待改成对象方法判断
                if api_result.message == '商品不存在' or api_result.message == '商品已下架':
                    return GetGoodDetailStatus(good_not_existed=True), None
                if isinstance(good, Good):
                    return GetGoodDetailStatus(success=True), good.update(api_result.data)
                else:
                    return GetGoodDetailStatus(success=True), Good.parse_obj(api_result.data)
    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"米游币商品兑换 - 获取商品详细信息: 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return GetGoodDetailStatus(incorrect_return=True), None
        else:
            logger.exception(f"米游币商品兑换 - 获取商品详细信息: 网络请求失败")
            return GetGoodDetailStatus(network_error=True), None


async def get_good_games(retry: bool = True) -> Tuple[BaseApiStatus, Optional[List[Tuple[str, str]]]]:
    """
    获取商品分区列表

    :param retry: 是否允许重试
    :return: (商品分区全名, 字母简称) 的列表
    """
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    res = await client.get(URL_GOOD_LIST.format(page=1,
                                                                game=""),
                                           headers=HEADERS_GOOD_LIST,
                                           timeout=conf.preference.timeout)
                api_result = ApiResultHandler(res.json())
                return BaseApiStatus(success=True), list(map(lambda x: (x["name"], x["key"]), api_result.data["games"]))
    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"米游币商品兑换 - 获取商品列表: 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return BaseApiStatus(incorrect_return=True), None
        else:
            logger.exception(f"米游币商品兑换 - 获取商品列表: 网络请求失败")
            return BaseApiStatus(network_error=True), None


async def get_good_list(game: str = "", retry: bool = True) -> Tuple[
    BaseApiStatus, Optional[List[Good]]]:
    """
    获取商品信息列表

    :param game: 游戏简称（默认为空，即获取所有游戏的商品）
    :param retry: 是否允许重试
    :return: 商品信息列表
    """
    good_list = []
    page = 1

    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    res = await client.get(URL_GOOD_LIST.format(page=page,
                                                                game=game), headers=HEADERS_GOOD_LIST,
                                           timeout=conf.preference.timeout)
                api_result = ApiResultHandler(res.json())
                goods = map(Good.parse_obj, api_result.data["list"])
                # 判断是否已经读完所有商品
                if not goods:
                    break
                else:
                    good_list += goods
                page += 1
    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"米游币商品兑换 - 获取商品列表: 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return BaseApiStatus(incorrect_return=True), None
        else:
            logger.exception(f"米游币商品兑换 - 获取商品列表: 网络请求失败")
            return BaseApiStatus(network_error=True), None

    return BaseApiStatus(success=True), good_list


async def get_address(account: UserAccount, retry: bool = True) -> Tuple[BaseApiStatus, Optional[List[Address]]]:
    """
    获取用户的地址数据

    :param account: 用户账户数据
    :param retry: 是否允许重试
    """
    headers = HEADERS_ADDRESS.copy()
    headers["x-rpc-device_id"] = account.device_id_ios
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    res = await client.get(
                        URL_ADDRESS.format(round(NtpTime.time() * 1000)),
                        headers=headers,
                        cookies=account.cookies.dict(v2_stoken=True, cookie_type=True),
                        timeout=conf.preference.timeout
                    )
                    api_result = ApiResultHandler(res.json())
                    if api_result.login_expired:
                        logger.info(
                            f"获取地址数据 - 用户 {account.bbs_uid} 登录失效")
                        logger.debug(f"网络请求返回: {res.text}")
                        return BaseApiStatus(login_expired=True), None
                address_list = list(map(Address.parse_obj, api_result.data["list"]))
    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"获取地址数据 - 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return BaseApiStatus(incorrect_return=True), None
        else:
            logger.exception(f"获取地址数据 - 请求失败")
            return BaseApiStatus(network_error=True), None
    return BaseApiStatus(success=True), address_list


async def check_registrable(phone_number: int, keep_client: bool = False, retry: bool = True) -> Tuple[
    BaseApiStatus, Optional[bool], str, Optional[httpx.AsyncClient]]:
    """
    检查用户是否可以注册

    :param keep_client: httpx.AsyncClient 连接是否需要关闭
    :param phone_number: 手机号
    :param retry: 是否允许重试
    :return: (API返回状态, 用户是否可以注册, 设备ID, httpx.AsyncClient连接对象)
    """
    headers = HEADERS_WEBAPI.copy()
    device_id = generate_device_id()
    headers["x-rpc-device_id"] = device_id

    async def request():
        """
        发送请求的闭包函数
        """
        time_now = round(NtpTime.time() * 1000)
        # await client.options(URL_REGISTRABLE.format(mobile=phone_number, t=time_now),
        #                      headers=headers, timeout=conf.preference.timeout)
        return await client.get(URL_REGISTRABLE.format(mobile=phone_number, t=time_now),
                                headers=headers, timeout=conf.preference.timeout)

    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                if keep_client:
                    client = httpx.AsyncClient()
                else:
                    async with httpx.AsyncClient() as client:
                        res = await request()
                res = await request()
                api_result = ApiResultHandler(res.json())
                return BaseApiStatus(success=True), bool(api_result.data["is_registable"]), device_id, client
    except tenacity.RetryError as e:
        if keep_client:
            await client.aclose()
        if is_incorrect_return(e):
            logger.exception(f"检查用户 {phone_number} 是否可以注册 - 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return BaseApiStatus(incorrect_return=True), None, device_id, client
        else:
            logger.exception(f"检查用户 {phone_number} 是否可以注册 - 请求失败")
            return BaseApiStatus(network_error=True), None, device_id, None


async def create_mmt(client: Optional[httpx.AsyncClient] = None,
                     use_v4: bool = True,
                     device_id: str = None,
                     retry: bool = True) -> Tuple[
    BaseApiStatus, Optional[MmtData], str, Optional[httpx.AsyncClient]]:
    """
    发送短信验证前所需的人机验证任务申请

    :param client: httpx.AsyncClient 连接
    :param use_v4: 是否使用极验第四代人机验证
    :param device_id: 设备ID
    :param retry: 是否允许重试
    :return: (API返回状态, 人机验证任务数据, 设备ID, httpx.AsyncClient连接对象)
    """
    headers = HEADERS_WEBAPI.copy()
    headers["x-rpc-device_id"] = device_id or generate_device_id()
    if use_v4:
        headers.setdefault("x-rpc-source", "accountWebsite")

    async def request():
        """
        发送请求的闭包函数
        """
        time_now = round(NtpTime.time() * 1000)
        # await client.options(URL_CREATE_MMT.format(now=time_now, t=time_now),
        #                      headers=headers, timeout=conf.preference.timeout)
        return await client.get(URL_CREATE_MMT.format(now=time_now, t=time_now),
                                headers=headers, timeout=conf.preference.timeout)

    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                if client:
                    res = await request()
                else:
                    async with httpx.AsyncClient() as client:
                        res = await request()
                api_result = ApiResultHandler(res.json())
                return BaseApiStatus(success=True), MmtData.parse_obj(api_result.data["mmt_data"]), device_id, client
    except tenacity.RetryError as e:
        if client:
            await client.aclose()
        if is_incorrect_return(e):
            logger.exception(f"获取短信验证-人机验证任务(create_mmt) - 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return BaseApiStatus(incorrect_return=True), None, device_id, client
        else:
            logger.exception(f"获取短信验证-人机验证任务(create_mmt) - 请求失败")
            return BaseApiStatus(network_error=True), None, device_id, None


async def create_mobile_captcha(phone_number: int,
                                mmt_data: MmtData,
                                geetest_result: Union[GeetestResult, GeetestResultV4],
                                client: Optional[httpx.AsyncClient] = None,
                                use_v4: bool = True,
                                device_id: str = None,
                                retry: bool = True
                                ) -> Tuple[CreateMobileCaptchaStatus, Optional[httpx.AsyncClient]]:
    """
    发送短信验证码

    :param phone_number: 手机号
    :param mmt_data: 人机验证任务数据
    :param geetest_result: 人机验证结果数据
    :param client: httpx.AsyncClient 连接
    :param use_v4: 是否使用极验第四代人机验证
    :param device_id: 设备ID
    :param retry: 是否允许重试
    """
    headers = HEADERS_WEBAPI.copy()
    headers["x-rpc-device_id"] = device_id or generate_device_id()
    if use_v4 and isinstance(geetest_result, GeetestResultV4):
        geetest_v4_data = geetest_result.dict(skip_defaults=True)
        content = {
            "action_type": "login",
            "mmt_key": mmt_data.mmt_key,
            "geetest_v4_data": str(geetest_v4_data).replace("'", '"'),
            "mobile": str(phone_number),
            "t": str(round(NtpTime.time() * 1000))
        }
    else:
        content = {
            "action_type": "login",
            "mmt_key": mmt_data.mmt_key,
            "geetest_challenge": mmt_data.challenge,
            "geetest_validate": geetest_result.validate,
            "geetest_seccode": geetest_result.seccode,
            "mobile": phone_number,
            "t": round(NtpTime.time() * 1000)
        }

    async def request():
        """
        发送请求的闭包函数
        """
        return await client.post(URL_CREATE_MOBILE_CAPTCHA,
                                 data=content,
                                 headers=headers,
                                 timeout=conf.preference.timeout)

    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                if client and not client.is_closed:
                    res = await request()
                else:
                    async with httpx.AsyncClient() as client:
                        res = await request()
                api_result = ApiResultHandler(res.json())
                if api_result.success:
                    return CreateMobileCaptchaStatus(success=True), client
                elif api_result.wrong_captcha:
                    return CreateMobileCaptchaStatus(incorrect_geetest=True), client
    except tenacity.RetryError as e:
        if client:
            await client.aclose()
        if is_incorrect_return(e):
            logger.exception(f"发送短信验证码 - 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return CreateMobileCaptchaStatus(incorrect_return=True), client
        else:
            logger.exception(f"发送短信验证码 - 请求失败")
            return CreateMobileCaptchaStatus(network_error=True), None


async def get_login_ticket_by_captcha(phone_number: str,
                                      captcha: int,
                                      client: Optional[httpx.AsyncClient] = None,
                                      device_id: str = None,
                                      retry: bool = True) -> \
        Tuple[
            GetCookieStatus, Optional[BBSCookies]]:
    """
    通过短信验证码获取 login_ticket

    :param phone_number: 手机号
    :param captcha: 短信验证码
    :param client: httpx.AsyncClient 连接
    :param device_id: 设备ID
    :param retry: 是否允许重试

    >>> import asyncio
    >>> coroutine = get_cookie_token_by_captcha("12345678910", 123456)
    >>> assert asyncio.new_event_loop().run_until_complete(coroutine)[0].incorrect_captcha is True
    """

    headers = HEADERS_WEBAPI.copy()
    headers["x-rpc-device_id"] = device_id or generate_device_id()
    params = {
        "mobile": phone_number,
        "mobile_captcha": captcha,
        "source": "user.mihoyo.com",
        "t": round(NtpTime.time() * 1000),
    }
    encoded_params = urlencode(params)

    async def request():
        """
        发送请求的闭包函数
        """
        # TODO 还需要进一步简化代码
        return await client.post(URL_LOGIN_TICKET_BY_CAPTCHA,
                                 headers=headers,
                                 content=encoded_params,
                                 timeout=conf.preference.timeout
                                 )

    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                if client is not None:
                    res = await request()
                else:
                    async with httpx.AsyncClient() as client:
                        res = await request()
                api_result = ApiResultHandler(res.json())
                if api_result.success:
                    cookies = BBSCookies.parse_obj(dict_from_cookiejar(
                        res.cookies.jar))
                    if not cookies.login_ticket:
                        return GetCookieStatus(missing_login_ticket=True), None
                    else:
                        if client:
                            await client.aclose()
                        return GetCookieStatus(success=True), cookies
                elif api_result.wrong_captcha:
                    logger.info(
                        f"通过短信验证码获取 login_ticket - 验证码错误，但你可以再次尝试登录")
                    return GetCookieStatus(incorrect_captcha=True), None
                else:
                    raise IncorrectReturn
    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"通过短信验证码获取 login_ticket: 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return GetCookieStatus(incorrect_return=True), None
        else:
            logger.exception(f"通过短信验证码获取 login_ticket: 网络请求失败")
            return GetCookieStatus(network_error=True), None


async def get_multi_token_by_login_ticket(cookies: BBSCookies, retry: bool = True) -> Tuple[
    GetCookieStatus, Optional[BBSCookies]]:
    """
    通过 login_ticket 获取 `stoken 和 ltoken

    :param cookies: 米游社Cookies，需要包含 login_ticket 和 bbs_uid
    :param retry: 是否允许重试
    """
    if not cookies.login_ticket:
        return GetCookieStatus(missing_login_ticket=True), None
    elif not cookies.bbs_uid:
        return GetCookieStatus(missing_bbs_uid=True), None
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    res = await client.get(
                        URL_MULTI_TOKEN_BY_LOGIN_TICKET.format(cookies.login_ticket, cookies.bbs_uid),
                        headers=HEADERS_API_TAKUMI_PC,
                        timeout=conf.preference.timeout)
                api_result = ApiResultHandler(res.json())
                if api_result.login_expired:
                    logger.warning(f"通过 login_ticket 获取 stoken: 登录失效")
                    return GetCookieStatus(login_expired=True), None
                else:
                    cookies.stoken = list(filter(
                        lambda x: x["name"] == "stoken", api_result.data["list"]))[0]["token"]
                    cookies.ltoken = list(filter(
                        lambda x: x["name"] == "ltoken", api_result.data["list"]))[0]["token"]
                    return GetCookieStatus(success=True), cookies
    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"通过 login_ticket 获取 stoken: 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return GetCookieStatus(incorrect_return=True), None
        else:
            logger.exception(f"通过 login_ticket 获取 stoken: 网络请求失败")
            return GetCookieStatus(network_error=True), None


async def get_cookie_token_by_captcha(phone_number: str, captcha: int, retry: bool = True) -> Tuple[
    GetCookieStatus, Optional[BBSCookies]]:
    """
    通过短信验证码获取 cookie_token

    :param phone_number: 手机号
    :param captcha: 验证码
    :param retry: 是否允许重试

    >>> import asyncio
    >>> coroutine = get_cookie_token_by_captcha("12345678910", 123456)
    >>> assert asyncio.new_event_loop().run_until_complete(coroutine)[0].incorrect_captcha is True
    """
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    res = await client.post(URL_COOKIE_TOKEN_BY_CAPTCHA,
                                            headers=HEADERS_API_TAKUMI_PC,
                                            json={
                                                "is_bh2": False,
                                                "mobile": phone_number,
                                                "captcha": str(captcha),
                                                "action_type": "login",
                                                "token_type": 6
                                            },
                                            timeout=conf.preference.timeout
                                            )
                api_result = ApiResultHandler(res.json())
                if api_result.wrong_captcha:
                    logger.info(f"登录米哈游账号 - 验证码错误")
                    return GetCookieStatus(incorrect_captcha=True), None
                else:
                    cookies = BBSCookies.parse_obj(dict_from_cookiejar(res.cookies.jar))
                    if not cookies.cookie_token:
                        return GetCookieStatus(missing_cookie_token=True), None
                    elif not cookies.bbs_uid:
                        return GetCookieStatus(missing_bbs_uid=True), None
                    else:
                        return GetCookieStatus(success=True), cookies
    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"通过短信验证码获取 cookie_token: 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return GetCookieStatus(incorrect_return=True), None
        else:
            logger.exception(f"通过短信验证码获取 cookie_token: 网络请求失败")
            return GetCookieStatus(network_error=True), None


async def get_login_ticket_by_password(account: str,
                                       password: str,
                                       mmt_data: MmtData,
                                       geetest_result: GeetestResult,
                                       device_id: str = None,
                                       retry: bool = True) -> Tuple[GetCookieStatus, Optional[BBSCookies]]:
    """
    使用密码登录获取login_ticket

    :param account: 账号
    :param password: 密码
    :param mmt_data: GEETEST验证任务数据
    :param geetest_result: GEETEST验证结果数据
    :param device_id: 设备ID
    :param retry: 是否允许重试
    """
    headers = HEADERS_WEBAPI.copy()
    headers["x-rpc-device_id"] = device_id or generate_device_id()
    params = {
        "account": account,
        "password": password,
        "is_crypto": False,
        "mmt_key": mmt_data.mmt_key,
        "geetest_challenge": mmt_data.challenge,
        "geetest_validate": geetest_result.validate,
        "geetest_seccode": geetest_result.seccode,
        "source": "user.mihoyo.com",
        "t": round(NtpTime.time() * 1000)
    }
    encoded_params = urlencode(params)
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    res = await client.post(
                        URL_LOGIN_TICKET_BY_PASSWORD,
                        content=encoded_params,
                        headers=headers,
                        timeout=conf.preference.timeout
                    )
                cookies = BBSCookies.parse_obj(dict_from_cookiejar(res.cookies.jar))
                api_result = ApiResultHandler(res.json())
                if api_result.success:
                    return GetCookieStatus(success=True), cookies
                elif api_result.wrong_captcha:
                    logger.warning(f"使用密码登录获取login_ticket - 图片验证码失败")
                    return GetCookieStatus(incorrect_captcha=True), None
    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"使用密码登录获取login_ticket - 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return GetCookieStatus(incorrect_return=True), None
        else:
            logger.exception(f"使用密码登录获取login_ticket - 请求失败")
            return GetCookieStatus(network_error=True), None


async def get_cookie_token_by_stoken(cookies: BBSCookies, device_id: str = None, retry: bool = True) -> Tuple[
    GetCookieStatus, Optional[BBSCookies]]:
    """
    通过 stoken_v2 获取 cookie_token

    :param cookies: 米游社Cookies，需要包含 stoken_v2 和 mid
    :param device_id: X_RPC_DEVICE_ID
    :param retry: 是否允许重试

    >>> import asyncio
    >>> coroutine = get_cookie_token_by_stoken(BBSCookies())
    >>> assert asyncio.new_event_loop().run_until_complete(coroutine)[0].success is False
    """
    headers = HEADERS_PASSPORT_API.copy()
    headers["x-rpc-device_id"] = device_id or generate_device_id()
    if not cookies.stoken_v2:
        return GetCookieStatus(missing_stoken_v2=True), None
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    res = await client.get(
                        URL_COOKIE_TOKEN_BY_STOKEN,
                        cookies=cookies.dict(v2_stoken=True, cookie_type=True),
                        headers=headers,
                        timeout=conf.preference.timeout
                    )
                api_result = ApiResultHandler(res.json())
                if api_result.success:
                    cookies.cookie_token = api_result.data["cookie_token"]
                    if not cookies.bbs_uid:
                        cookies.bbs_uid = api_result.data["uid"]
                    return GetCookieStatus(success=True), cookies
                elif api_result.login_expired:
                    logger.warning(f"通过 stoken 获取 cookie_token: 登录失效")
                    return GetCookieStatus(login_expired=True), None
                else:
                    raise IncorrectReturn

    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"通过 stoken 获取 cookie_token: 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return GetCookieStatus(incorrect_return=True), None
        else:
            logger.exception(f"通过 stoken 获取 cookie_token: 网络请求失败")
            return GetCookieStatus(network_error=True), None


async def get_stoken_v2_by_v1(cookies: BBSCookies, device_id: str = None, retry: bool = True) -> Tuple[
    GetCookieStatus, Optional[BBSCookies]]:
    """
    通过 stoken_v1 获取 stoken_v2 以及 mid

    :param cookies: 米游社Cookies，需要包含 stoken_v1
    :param device_id: X_RPC_DEVICE_ID
    :param retry: 是否允许重试

    >>> import asyncio
    >>> coroutine = get_stoken_v2_by_v1(BBSCookies())
    >>> assert asyncio.new_event_loop().run_until_complete(coroutine)[0].success is False
    """
    headers = HEADERS_PASSPORT_API.copy()
    headers["x-rpc-device_id"] = device_id or generate_device_id()
    headers.setdefault("x-rpc-aigis", "")
    headers.setdefault("x-rpc-app_id", "bll8iq97cem8")

    if not cookies.stoken_v1:
        return GetCookieStatus(missing_stoken_v1=True), None
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    headers.setdefault("DS", generate_ds(salt=conf.salt_config.SALT_PROD))
                    res = await client.post(
                        URL_STOKEN_V2_BY_V1,
                        cookies={"stoken": cookies.stoken_v1, "stuid": cookies.bbs_uid},
                        headers=headers,
                        timeout=conf.preference.timeout
                    )
                api_result = ApiResultHandler(res.json())
                if api_result.success:
                    cookies.stoken_v2 = api_result.data["token"]["token"]
                    cookies.mid = api_result.data["user_info"]["mid"]
                    if not cookies.bbs_uid:
                        cookies.bbs_uid = api_result.data["user_info"]["aid"]
                    return GetCookieStatus(success=True), cookies
                elif api_result.login_expired:
                    logger.warning(f"通过 stoken_v1 获取 stoken_v2: 登录失效")
                    return GetCookieStatus(login_expired=True), None
                else:
                    raise IncorrectReturn

    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"通过 stoken_v1 获取 stoken_v2: 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return GetCookieStatus(incorrect_return=True), None
        else:
            logger.exception(f"通过 stoken_v1 获取 stoken_v2: 网络请求失败")
            return GetCookieStatus(network_error=True), None


async def get_ltoken_by_stoken(cookies: BBSCookies, device_id: str = None, retry: bool = True) -> Tuple[
    GetCookieStatus, Optional[BBSCookies]]:
    """
    通过 stoken_v2 和 mid 获取 ltoken

    :param cookies: 米游社Cookies，需要包含 stoken_v2 和 mid
    :param device_id: X_RPC_DEVICE_ID
    :param retry: 是否允许重试

    >>> import asyncio
    >>> coroutine = get_ltoken_by_stoken(BBSCookies())
    >>> assert asyncio.new_event_loop().run_until_complete(coroutine)[0].success is False
    """
    headers = HEADERS_PASSPORT_API.copy()
    headers["x-rpc-device_id"] = device_id or generate_device_id()
    if not cookies.stoken_v2:
        return GetCookieStatus(missing_stoken_v2=True), None
    if not cookies.mid:
        return GetCookieStatus(missing_mid=True), None
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    res = await client.get(
                        URL_LTOKEN_BY_STOKEN,
                        cookies=cookies.dict(v2_stoken=True, cookie_type=True),
                        headers=headers,
                        timeout=conf.preference.timeout
                    )
                api_result = ApiResultHandler(res.json())
                if api_result.success:
                    cookies.ltoken = api_result.data["ltoken"]
                    return GetCookieStatus(success=True), cookies
                elif api_result.login_expired:
                    logger.warning(f"通过 stoken 获取 ltoken: 登录失效")
                    return GetCookieStatus(login_expired=True), None
                else:
                    raise IncorrectReturn

    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"通过 stoken 获取 ltoken: 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return GetCookieStatus(incorrect_return=True), None
        else:
            logger.exception(f"通过 stoken 获取 ltoken: 网络请求失败")
            return GetCookieStatus(network_error=True), None


async def get_device_fp(device_id: str, retry: bool = True) -> Tuple[GetFpStatus, Optional[str]]:
    """
    获取 x-rpc-device_fp

    :param device_id: x-rpc-device_id 的值
    :param retry: 是否允许重试

    >>> import asyncio
    >>> coroutine = get_device_fp(generate_device_id())
    >>> assert asyncio.new_event_loop().run_until_complete(coroutine)[0].success is True
    """
    content = {
        "seed_id": generate_seed_id(),
        "device_id": device_id.lower(),
        "platform": "5",
        "seed_time": str(int(time.time() * 1000)),
        "ext_fields": "{\"userAgent\":\"Mozilla\/5.0 (iPhone; CPU iPhone OS 16_2 like Mac OS X) AppleWebKit\/605.1.15 "
                      f"(KHTML, like Gecko) miHoYoBBS\/{conf.device_config.X_RPC_APP_VERSION}\",\"browserScreenSize"
                      f"\":243750,\"maxTouchPoints\":5,"
                      "\"isTouchSupported\":true,\"browserLanguage\":\"zh-CN\",\"browserPlat\":\"iPhone\","
                      "\"browserTimeZone\":\"Asia\/Shanghai\",\"webGlRender\":\"Apple GPU\",\"webGlVendor\":\"Apple "
                      "Inc.\",\"numOfPlugins\":0,\"listOfPlugins\":\"unknown\",\"screenRatio\":3,"
                      "\"deviceMemory\":\"unknown\",\"hardwareConcurrency\":\"4\",\"cpuClass\":\"unknown\","
                      "\"ifNotTrack\":\"unknown\",\"ifAdBlock\":0,\"hasLiedResolution\":1,\"hasLiedOs\":0,"
                      "\"hasLiedBrowser\":0}",
        "app_name": "account_cn",
        "device_fp": generate_fp_locally()
    }
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    res = await client.post(
                        URL_GET_DEVICE_FP,
                        json=content,
                        timeout=conf.preference.timeout
                    )
                api_result = ApiResultHandler(res.json())
                if api_result.data["code"] == 403 or api_result.data["msg"] == "传入的参数有误":
                    logger.error("传入的参数有误")
                    return GetFpStatus(invalid_arguments=True), None
                elif api_result.success:
                    device_fp = api_result.data["device_fp"]
                    if not device_fp:
                        logger.error("获取 x-rpc-device_fp: 服务器返回的 device_fp 为空")
                        return GetFpStatus(incorrect_return=True), None
                    return GetFpStatus(success=True), device_fp

    except tenacity.RetryError as e:
        if is_incorrect_return(e):
            logger.exception(f"获取 x-rpc-device_fp: 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return GetFpStatus(incorrect_return=True), None
        else:
            logger.exception(f"获取 x-rpc-device_fp: 网络请求失败")
            return GetFpStatus(network_error=True), None


async def good_exchange(plan: ExchangePlan) -> Tuple[ExchangeStatus, Optional[ExchangeResult]]:
    """
    执行米游币商品兑换

    :param plan: 兑换计划
    """
    headers = HEADERS_EXCHANGE
    headers["x-rpc-device_id"] = plan.account.device_id_ios
    headers["x-rpc-device_fp"] = plan.account.device_fp or generate_fp_locally()
    content = {
        "app_id": 1,
        "point_sn": "myb",
        "goods_id": plan.good.goods_id,
        "exchange_num": 1
    }
    if plan.address is not None:
        content.setdefault("address_id", plan.address.id)
    if plan.game_record is not None:
        content.setdefault("uid", plan.game_record.game_role_id)
        # 例: cn_gf01
        content.setdefault("region", plan.game_record.region)
        # 例: hk4e_cn
        content.setdefault("game_biz", plan.good.game_biz)
    try:
        async with httpx.AsyncClient() as client:
            res = await client.post(
                URL_EXCHANGE, headers=headers, json=content,
                cookies=plan.account.cookies.dict(cookie_type=True),
                timeout=conf.preference.timeout)
        api_result = ApiResultHandler(res.json())
        if api_result.login_expired:
            logger.info(
                f"米游币商品兑换 - 执行兑换: 用户 {plan.account.bbs_uid} 登录失效")
            logger.debug(f"网络请求返回: {res.text}")
            return ExchangeStatus(login_expired=True), None
        if api_result.success:
            logger.info(
                f"米游币商品兑换 - 执行兑换: 用户 {plan.account.bbs_uid} 商品 {plan.good.goods_id} 兑换成功！可以自行确认。")
            logger.debug(f"网络请求返回: {res.text}")
            return ExchangeStatus(success=True), ExchangeResult(result=True, return_data=res.json(), plan=plan)
        else:
            logger.info(
                f"米游币商品兑换 - 执行兑换: 用户 {plan.account.bbs_uid} 商品 {plan.good.goods_id} 兑换失败，可以自行确认。")
            logger.debug(f"网络请求返回: {res.text}")
            return ExchangeStatus(success=True), ExchangeResult(result=False, return_data=res.json(), plan=plan)
    except Exception as e:
        if is_incorrect_return(e):
            logger.error(
                f"米游币商品兑换 - 执行兑换: 用户 {plan.account.bbs_uid} 商品 {plan.good.goods_id} 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return ExchangeStatus(incorrect_return=True), None
        else:
            logger.exception(
                f"米游币商品兑换 - 执行兑换: 用户 {plan.account.bbs_uid} 商品 {plan.good.goods_id} 请求失败")
            return ExchangeStatus(network_error=True), None


def good_exchange_sync(plan: ExchangePlan) -> Tuple[ExchangeStatus, Optional[ExchangeResult]]:
    """
    执行米游币商品兑换

    :param plan: 兑换计划
    """
    headers = HEADERS_EXCHANGE
    headers["x-rpc-device_id"] = plan.account.device_id_ios
    headers["x-rpc-device_fp"] = plan.account.device_fp or generate_fp_locally()
    content = {
        "app_id": 1,
        "point_sn": "myb",
        "goods_id": plan.good.goods_id,
        "exchange_num": 1
    }
    if plan.address is not None:
        content.setdefault("address_id", plan.address.id)
    if plan.game_record is not None:
        content.setdefault("uid", plan.game_record.game_role_id)
        # 例: cn_gf01
        content.setdefault("region", plan.game_record.region)
        # 例: hk4e_cn
        content.setdefault("game_biz", plan.good.game_biz)
    try:
        with httpx.Client() as client:
            res = client.post(
                URL_EXCHANGE, headers=headers, json=content,
                cookies=plan.account.cookies.dict(cookie_type=True),
                timeout=conf.preference.timeout)
        api_result = ApiResultHandler(res.json())
        if api_result.login_expired:
            logger.info(
                f"米游币商品兑换 - 执行兑换: 用户 {plan.account.bbs_uid} 登录失效")
            logger.debug(f"网络请求返回: {res.text}")
            return ExchangeStatus(login_expired=True), None
        if api_result.success:
            logger.info(
                f"米游币商品兑换 - 执行兑换: 用户 {plan.account.bbs_uid} 商品 {plan.good.goods_id} 兑换成功！可以自行确认。")
            logger.debug(f"网络请求返回: {res.text}")
            return ExchangeStatus(success=True), ExchangeResult(result=True, return_data=res.json(), plan=plan)
        else:
            logger.info(
                f"米游币商品兑换 - 执行兑换: 用户 {plan.account.bbs_uid} 商品 {plan.good.goods_id} 兑换失败，可以自行确认。")
            logger.debug(f"网络请求返回: {res.text}")
            return ExchangeStatus(success=True), ExchangeResult(result=False, return_data=res.json(), plan=plan)
    except Exception as e:
        if is_incorrect_return(e):
            logger.error(
                f"米游币商品兑换 - 执行兑换: 用户 {plan.account.bbs_uid} 商品 {plan.good.goods_id} 服务器没有正确返回")
            logger.debug(f"网络请求返回: {res.text}")
            return ExchangeStatus(incorrect_return=True), None
        else:
            logger.exception(
                f"米游币商品兑换 - 执行兑换: 用户 {plan.account.bbs_uid} 商品 {plan.good.goods_id} 请求失败")
            return ExchangeStatus(network_error=True), None



import asyncio
import json
import os
import threading
import traceback
from http.server import HTTPServer, ThreadingHTTPServer, CGIHTTPRequestHandler
from json import JSONDecodeError
from multiprocessing import Pipe, Manager, connection
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import Any, Optional, Callable, Tuple, Union
from urllib import parse

from pydantic import ValidationError

from mys_goods_tool.data_model import GeetestResult, GeetestResultV4
from mys_goods_tool.user_data import config as conf
from mys_goods_tool.utils import logger, get_free_port, ProcessManager

STATIC_DIRECTORY = Path(__file__).resolve().with_name(
    "geetest-webui") if not conf.preference.geetest_statics_path else conf.preference.geetest_statics_path


class GeetestHandler(CGIHTTPRequestHandler):
    result_callback: Callable[[Union[GeetestResult, GeetestResultV4]], Any]
    """接收验证结果数据的回调函数"""
    error_content_type = "application/json"

    def log_message(self, format: str, *args: Any) -> None:
        """
        重写父类方法，更改日志输出到loguru

        Log an arbitrary message.

        This is used by all other logging functions.  Override
        it if you have specific logging wishes.

        The first argument, FORMAT, is a format string for the
        message to be logged.  If the format string contains
        any % escapes requiring parameters, they should be
        specified as subsequent arguments (it's just like
        printf!).

        The client ip and current date/time are prefixed to
        every message.
        """
        logger.debug(f"HttpServer: {self.address_string()} - - {format % args}")

    def response_ok(self):
        response = {"code": 200, "message": "OK. Server received"}
        self.send_response(200, response["message"])
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps(response).encode())

    def send_error(self, code: int, message: str = None, explain: str = None, to_json: bool = False) -> None:
        """
        重写父类方法，返回json格式的错误信息
        """
        if to_json:
            self.send_response(code, message)
            self.send_header("Content-Type", self.error_content_type)
            self.end_headers()
            response = {"code": code, "msg": message, "desc": explain}
            self.wfile.write(json.dumps(response).encode())
        else:
            super().send_error(code, message, explain)

    def do_GET(self):
        file_path = self.path
        try:
            if self.path == "/" or not self.path:
                # 重定向到默认文件
                self.path = "/index.html"
            elif self.path.startswith("/result"):
                # 接收验证结果
                params = parse.parse_qs(parse.urlparse(self.path).query)
                seccode = params.get("seccode")
                validate = params.get("validate")
                if not seccode or not validate:
                    logger.error(
                        f"HTTP服务器 - 收到 {self.address_string()} 的验证结果，但URL参数缺少 seccode 和 validate")
                    self.send_error(400, "Bad request, missing URL params `seccode` and `validate`", to_json=True)
                else:
                    geetest_result = GeetestResult(seccode=seccode[0], validate=validate[0])
                    logger.info(f"HTTP服务器 - 收到 {self.address_string()} 的验证结果 - {geetest_result}")
                    self.result_callback(geetest_result)
                    self.response_ok()
                return

            # 构建文件路径
            file_path = STATIC_DIRECTORY / parse.urlparse(self.path).path[1:]

            # 检查文件是否存在并且可读
            if not os.path.isfile(file_path):
                raise FileNotFoundError(f"文件 {file_path} 不存在")
            if not os.access(file_path, os.R_OK):
                raise FileNotFoundError(f"文件 {file_path} 不可读")

            # 发送响应头
            self.send_response(200)
            self.send_header("Content-Type", self.guess_type(file_path))
            self.end_headers()

            # 发送文件内容
            with open(file_path, "rb") as file:
                self.wfile.write(file.read())
        except FileNotFoundError:
            logger.warning(f"HTTP服务器 - 收到 {self.address_string()} 的请求，但请求文件 {file_path} 不存在或不可读")
            logger.debug(traceback.format_exc())
            self.send_error(404)
        except Exception:
            logger.error(f"HTTP服务器 - 收到 {self.address_string()} 的请求，但处理请求时发生错误")
            logger.debug(traceback.format_exc())
            self.send_error(500)

    def do_POST(self) -> None:
        if self.path.startswith("/result"):
            try:
                content_length = int(self.headers["Content-Length"])
                geetest_result = GeetestResultV4.parse_raw(self.rfile.read(content_length))
            except (ValidationError, JSONDecodeError):
                logger.exception(
                    f"HTTP服务器 - 收到 {self.address_string()} 的 gt4 验证结果，但传入数据不符合 GeetestResultV4 模型")
                self.send_error(400, "Bad request, data not match GeetestResultV4 model", to_json=True)
            else:
                logger.info(f"HTTP服务器 - 收到 {self.address_string()} 的 gt4 验证结果")
                self.result_callback(geetest_result)
                self.response_ok()


def set_listen_address():
    """
    设置HTTP服务器监听地址
    """
    if conf.preference.geetest_listen_address:
        if conf.preference.geetest_listen_address[1] == 0:
            port = get_free_port(conf.preference.geetest_listen_address[0])
            if port is None:
                return
            else:
                return conf.preference.geetest_listen_address[0], port
        else:
            return conf.preference.geetest_listen_address
    else:
        port = get_free_port()
        if port is None:
            return
        else:
            return "localhost", get_free_port()


class GeetestProcess:
    """
    GEETEST行为验证HTTP服务器 进程相关
    （可被pickle序列化的对象类型有限，给该类引入新的数据时需要注意）
    """
    httpd: ThreadingHTTPServer
    """GEETEST行为验证HTTP服务器实例"""
    pipe: Tuple[connection.Connection]
    """进程间通信管道，用于终止HTTP服务器"""
    geetest_result_queue: Queue[GeetestResult]
    """由Manager管理的验证结果数据"""
    wait_thread: Thread

    @classmethod
    def result_callback(cls, result: Union[GeetestResult, GeetestResultV4]):
        """
        给主进程发送验证结果
        """
        cls.geetest_result_queue.put(result)

    @classmethod
    def wait_until_stop(cls):
        connection.wait([cls.pipe[0]])
        cls.httpd.shutdown()
        logger.info(f"HTTP服务器 - 运行于 {cls.httpd.server_address} 的服务器已停止")

    @classmethod
    def run(cls, listen_address: Tuple[str, int], pipe: Tuple[connection.Connection],
            geetest_result_queue: Queue[GeetestResult]):
        """
        进程所执行的任务，被阻塞（启动HTTP服务器）

        :param listen_address HTTP服务器监听地址
        :param pipe 进程间通信管道，用于终止HTTP服务器
        :param geetest_result_queue 由Manager管理的验证结果数据
        """
        cls.pipe = pipe
        cls.geetest_result_queue = geetest_result_queue
        GeetestHandler.result_callback = cls.result_callback
        cls.httpd = ThreadingHTTPServer(listen_address, GeetestHandler)
        cls.wait_thread = threading.Thread(target=cls.wait_until_stop, daemon=True)
        cls.wait_thread.start()
        logger.info(
            f"HTTP服务器 - 监听地址：http://{listen_address[0]}:{listen_address[1]}")
        cls.httpd.serve_forever()


class GeetestProcessManager(ProcessManager):
    """
    异步GEETEST验证服务器，包含进程池对象
    """

    def __init__(self, listen_address: Tuple[str, int], httpd_close_callback: Optional[Callable] = None,
                 error_httpd_callback: Optional[Callable] = None):
        """
        创建进程池，初始化异步GEETEST验证服务器，包含进程池对象

        :param listen_address: HTTP服务器监听地址
        :param httpd_close_callback: HTTP服务器正常结束后的回调函数
        :param error_httpd_callback: HTTP服务器发生错误后的回调函数
        """
        self.listen_address = listen_address
        self.pipe: Tuple[connection.Connection] = Pipe(duplex=False)
        self.result_queue: Queue[GeetestResult] = Manager().Queue()

        super().__init__(httpd_close_callback, error_httpd_callback)

    def start(self, *_):
        """
        启动HTTP服务器
        """
        super().start(GeetestProcess.run, [self.listen_address, self.pipe, self.result_queue])

    async def force_stop_later(self, delay: float):
        """
        延迟一段时间后强制停止HTTP服务器
        """
        asyncio.get_event_loop().call_later(delay, self.pool.terminate)


class SetAddressProcessManager(ProcessManager):
    def start(self, *_):
        """
        开始寻找可用的地址
        """
        super().start(set_listen_address, [])


class GeetestServerThread(Thread):
    """
    后台启动GEETEST行为验证HTTP服务器的线程
    """

    def __init__(self):
        super().__init__()
        self.httpd: Optional[HTTPServer] = None

    def run(self):
        try:
            listen_address = set_listen_address()
            self.httpd = HTTPServer(listen_address, GeetestHandler)
            logger.info(
                f"HTTP服务器 - 监听地址：http://{listen_address[0]}:{listen_address[1]}")
            self.httpd.serve_forever()
        except Exception:
            logger.error("HTTP服务器 - 启动失败")
            logger.debug(traceback.format_exc())

    def stop_server(self):
        """
        关闭HTTP服务器
        """
        self.httpd.shutdown()
        logger.info(f"HTTP服务器 - 运行于 {self.httpd.server_address} 的服务器已停止")


def __get_result_demo(_: GeetestHandler, geetest_result: GeetestResult):
    """
    用于演示的验证结果处理函数

    :param _: GeetestHandler 实例
    :param geetest_result: 验证结果数据
    """
    logger.info(str(geetest_result))



from __future__ import annotations

import asyncio
import queue
from typing import NamedTuple, Tuple, Optional, Set
from urllib.parse import urlencode

from rich.markdown import Markdown
from textual.app import ComposeResult
from textual.widgets import (
    Input
)

from mys_goods_tool.api import create_mobile_captcha, create_mmt, get_login_ticket_by_captcha, \
    get_multi_token_by_login_ticket, get_cookie_token_by_stoken, get_stoken_v2_by_v1, get_ltoken_by_stoken, \
    get_device_fp
from mys_goods_tool.custom_css import *
from mys_goods_tool.custom_widget import RadioStatus, StaticStatus, ControllableButton, LoadingDisplay
from mys_goods_tool.data_model import GeetestResult, MmtData, GetCookieStatus
from mys_goods_tool.exchange_plan_view import ExchangePlanView
from mys_goods_tool.geetest import GeetestProcessManager, SetAddressProcessManager
from mys_goods_tool.user_data import config as conf, UserAccount, CONFIG_PATH
from mys_goods_tool.utils import logger


class LoginView(Container):
    """
    登录页面
    """
    DEFAULT_CSS = """
    LoginView {
        height: auto;
    }
    """

    def compose(self) -> ComposeResult:
        yield CaptchaLoginInformation()
        yield PhoneForm()
        yield CaptchaForm()


class CaptchaLoginInformation(Container):
    """
    短信验证登录页面的信息提示区域
    """
    DEFAULT_CSS = """
    CaptchaLoginInformation {
        height: auto;
        margin: 1 0;
        overflow: hidden;
    }

    CaptchaLoginInformation Horizontal {
        align: center middle;
    }
    """

    RadioTuple = NamedTuple("RadioTuple",
                            create_geetest=RadioStatus,
                            http_server=RadioStatus,
                            geetest_finished=RadioStatus,
                            create_captcha=RadioStatus,
                            login_ticket_by_captcha=RadioStatus,
                            multi_token_by_login_ticket=RadioStatus,
                            get_stoken_v2=RadioStatus,
                            get_ltoken_by_stoken=RadioStatus,
                            cookie_token_by_stoken=RadioStatus,
                            login_finished=RadioStatus
                            )

    StaticTuple = NamedTuple("StaticTuple",
                             geetest_title=Static,
                             geetest_text=StaticStatus
                             )

    radio_tuple = RadioTuple(
        create_geetest=RadioStatus("短信验证码 - 申请人机验证任务"),
        http_server=RadioStatus("开启人机验证网页服务器"),
        geetest_finished=RadioStatus("完成人机验证"),
        create_captcha=RadioStatus("发出短信验证码"),
        login_ticket_by_captcha=RadioStatus("通过验证码获取 login_ticket"),
        multi_token_by_login_ticket=RadioStatus("通过 login_ticket 获取 stoken"),
        get_stoken_v2=RadioStatus("获取 v2 版本 stoken 和 mid"),
        get_ltoken_by_stoken=RadioStatus("通过 stoken_v2 获取 ltoken"),
        cookie_token_by_stoken=RadioStatus("通过 stoken_v2 获取 cookie_token"),
        login_finished=RadioStatus("完成登录")
    )

    SAVE_TEXT = str(CONFIG_PATH)
    GEETEST_TEXT = "- 暂无需要完成的人机验证任务 -"

    static_tuple = StaticTuple(
        geetest_title=Static(Markdown("## GEETEST人机验证链接")),
        geetest_text=StaticStatus(GEETEST_TEXT)
    )

    radio_set = CaptchaStepSet(*radio_tuple)
    static_set = CaptchaTips(*static_tuple)

    def compose(self) -> ComposeResult:
        yield Horizontal(self.radio_set, self.static_set)


class PhoneForm(LoginForm):
    """
    手机号 表单
    """
    input = Input(placeholder="手机号", id="login_phone")
    """手机号输入框"""
    device_id: Optional[str] = None
    """人机验证过程的设备ID"""

    ButtonTuple = NamedTuple("ButtonTuple",
                             send=ControllableButton,
                             stop_geetest=ControllableButton,
                             success=ControllableButton,
                             error=ControllableButton)

    def __init__(self):
        super().__init__()

        self.mmt_data: Optional[MmtData] = None
        """GEETEST行为验证任务数据"""
        self.geetest_manager: Optional[GeetestProcessManager] = None
        """包含进程池的GEETEST验证HTTP服务器 进程管理器"""
        self.set_address_manager = SetAddressProcessManager(self.set_address_callback,
                                                            self.set_address_error_callback)
        """包含进程池的可用监听地址获取 进程管理器"""
        self.loop = asyncio.get_event_loop()
        """事件循环"""
        self.loop_tasks: Set[asyncio.Task] = set()
        """异步任务集合（保留其强引用）"""
        self.before_create_captcha = True
        """当前状态是否处于按下“发送短信验证码”按钮之前"""

        self.loading = LoadingDisplay()
        self.loading.hide()

        self.button = self.ButtonTuple(
            send=ControllableButton("发送短信验证码", variant="primary", id="create_captcha_send"),
            stop_geetest=ControllableButton("放弃人机验证", variant="warning", id="create_captcha_stop_geetest"),
            success=ControllableButton("完成", variant="success", id="create_captcha_success"),
            error=ControllableButton("返回", variant="error", id="create_captcha_error")
        )
        for button in self.button[1:]:
            button.hide()

    def compose(self) -> ComposeResult:
        yield Static("手机号", classes="label")
        yield self.input
        yield Static()
        yield from self.button
        yield Static()
        yield self.loading

    def close_create_captcha_send(self):
        """
        关闭发送短信验证码按钮
        """
        self.loading.hide()
        self.button.send.hide()
        self.button.send.enable()

    def httpd_error_callback(self, exception: BaseException):
        """
        GEETEST验证HTTP服务器启动失败时的回调函数
        """
        logger.error("用于Geetest验证的HTTP服务器启动失败")
        logger.debug(exception)
        CaptchaLoginInformation.radio_tuple.http_server.turn_off()
        self.button.stop_geetest.hide()
        self.button.error.show()

    async def listen_result(self):
        """
        等待GEETEST验证结果的异步任务
        """
        self.app.notice("请前往链接进行验证")
        while True:
            await asyncio.sleep(1)
            try:
                geetest_result: GeetestResult = self.geetest_manager.result_queue.get_nowait()
            except queue.Empty:
                continue
            else:
                logger.info(f"已收到Geetest验证结果数据，将发送验证码至 {self.input.value}")
                CaptchaLoginInformation.radio_tuple.geetest_finished.turn_on()
                self.loading.show()
                create_captcha_status, _ = await create_mobile_captcha(
                    int(self.input.value),
                    self.mmt_data,
                    geetest_result,
                    device_id=PhoneForm.device_id
                )
                if create_captcha_status:
                    self.loading.hide()
                    logger.info(f"短信验证码已发送至 {self.input.value}")
                    CaptchaLoginInformation.radio_tuple.create_captcha.turn_on()
                    CaptchaLoginInformation.static_tuple.geetest_text.change_text(CaptchaLoginInformation.GEETEST_TEXT,
                                                                                  "center")
                    self.button.success.show()
                    self.button.stop_geetest.hide()

                    self.geetest_manager.pipe[1].send(True)
                    await self.geetest_manager.force_stop_later(10)

                    self.app.notice("短信验证码已发送至 [green]" + self.input.value + "[/]")
                    break
                else:
                    self.loading.hide()
                    self.button.error.show()
                    self.button.stop_geetest.hide()
                    CaptchaLoginInformation.static_tuple.geetest_text.change_text(CaptchaLoginInformation.GEETEST_TEXT,
                                                                                  "center")
                    self.app.notice("[red]短信验证码发送失败[/]")

    def set_address_callback(self, address: Tuple[str, int]):
        """
        可用监听地址获取成功时的回调函数

        :param address: 返回的可用地址
        """
        if not address:
            self.close_create_captcha_send()
            self.button.error.show()
            return
        self.geetest_manager = GeetestProcessManager(address, error_httpd_callback=self.httpd_error_callback)
        logger.info(f"尝试在 http://{address[0]}:{address[1]} 上启动用于Geetest验证的HTTP服务器")
        self.geetest_manager.start()

        self.close_create_captcha_send()
        self.button.stop_geetest.show()
        CaptchaLoginInformation.radio_tuple.http_server.turn_on()

        task = self.loop.create_task(self.listen_result())
        self.loop_tasks.add(task)
        task.add_done_callback(self.loop_tasks.discard)

        params = {
            "gt": self.mmt_data.gt,
            "mmtKey": self.mmt_data.mmt_key,
            "riskType": self.mmt_data.risk_type
        }
        url_params = urlencode(params)
        link = f"http://{address[0]}:{address[1]}/index.html?{url_params}"
        link_localized = f"http://{address[0]}:{address[1]}/localized.html?{url_params}"
        CaptchaLoginInformation.static_tuple.geetest_text.change_text(
            renderable=f"\n- 请前往链接进行验证：\n"
                       f"[@click=app.open_link('{link}')]{link}[/]\n"
                       f"\n- 如果页面加载慢或者出错，尝试：\n"
                       f"[@click=app.open_link('{link_localized}')]{link_localized}[/]\n"
                       f"\n如果你在使用SSH或WSL，无法跳转链接，可前往日志文件复制验证链接",
            text_align="left")
        logger.info(f"请前往链接进行人机验证：{link}")
        logger.info(f"如果页面加载慢或者出错，尝试：{link_localized}")

    def set_address_error_callback(self, exception: BaseException):
        """
        可用监听地址获取失败时的回调函数
        """
        logger.error("尝试获取可用HTTP监听地址失败")
        logger.debug(exception)
        self.close_create_captcha_send()
        self.button.error.show()
        self.app.notice("[red]尝试获取可用HTTP监听地址失败！[/]")
        return

    async def create_captcha(self):
        """
        发送验证码的完整操作
        """
        if not self.before_create_captcha:
            return
        elif not self.input.value:
            self.app.notice("登录信息缺少 [bold red]手机号[/] ！")
            return
        self.before_create_captcha = False

        [i.turn_off() for i in CaptchaLoginInformation.radio_tuple]
        self.button.send.disable()
        self.loading.show()

        create_mmt_status, self.mmt_data, PhoneForm.device_id, _ = await create_mmt()
        if not create_mmt_status:
            self.close_create_captcha_send()
            self.button.error.show()
            self.app.notice("[red]获取Geetest行为验证任务数据失败！[/]")
            return
        else:
            logger.info(f"已成功获取Geetest行为验证任务数据 {self.mmt_data}")
            CaptchaLoginInformation.radio_tuple.create_geetest.turn_on()
            self.set_address_manager.start()

        return create_mmt_status

    async def on_input_submitted(self, _: Input.Submitted):
        await self.create_captcha()

    async def _on_button_pressed(self, event: ControllableButton.Pressed):
        if event.button.id == "create_captcha_send":
            # 按下“发送短信验证码”按钮时触发的事件

            await self.create_captcha()

        elif event.button.id == "create_captcha_stop_geetest":
            # 按下“放弃人机验证”按钮时触发的事件

            CaptchaLoginInformation.static_tuple.geetest_text.change_text(CaptchaLoginInformation.GEETEST_TEXT,
                                                                          "center")
            [i.turn_off() for i in CaptchaLoginInformation.radio_tuple]
            self.geetest_manager.pipe[1].send(True)
            self.button.stop_geetest.hide()
            self.button.send.show()
            self.before_create_captcha = True
            await self.geetest_manager.force_stop_later(10)

        elif event.button.id in ["create_captcha_success", "create_captcha_error"]:
            # 按下“完成（成功）”或“返回（出错）”按钮时触发的事件

            if event.button.id == "create_captcha_error":
                [i.turn_off() for i in CaptchaLoginInformation.radio_tuple]
            self.button.success.hide()
            self.button.error.hide()
            self.button.send.show()
            self.before_create_captcha = True


class CaptchaForm(LoginForm):
    """
    验证码 表单
    """
    ButtonTuple = NamedTuple("ButtonTuple", login=ControllableButton, success=ControllableButton,
                             error=ControllableButton)

    def __init__(self):
        super().__init__()
        self.login_result: Optional[GetCookieStatus] = None
        """登录操作返回值"""
        self.before_login: bool = True
        """当前状态是否在登录操作之前（不处于正在登录的状态）"""

        self.input = Input(placeholder="若发送验证码失败，也可前往米哈游通信证页面手动发送", id="login_captcha")

        self.loading = LoadingDisplay()
        self.loading.hide()

        self.button = self.ButtonTuple(
            login=ControllableButton("登录 / 刷新Cookies", variant="primary", id="login"),
            success=ControllableButton("完成", variant="success", id="login_success"),
            error=ControllableButton("返回", variant="error", id="login_error")
        )
        for i in self.button[1:]:
            i.hide()

    def compose(self) -> ComposeResult:
        yield Static("验证码", classes="label")
        yield self.input
        yield Static()
        yield from self.button
        yield Static()
        yield self.loading

    def close_login(self):
        self.button.login.hide()
        self.button.login.enable()

    async def login(self):
        """
        登录的完整操作
        """
        if not self.before_login:
            return
        elif not PhoneForm.input.value:
            self.app.notice("登录信息缺少 [bold red]手机号[/] ！")
            return
        elif not self.input.value.isdigit() and self.input.value:
            self.app.notice("登录信息 [bold red]验证码[/] 需要是数字或为空（刷新Cookies）！")
            return
        self.before_login = False

        self.button.login.disable()
        self.loading.show()

        account: Optional[UserAccount] = None
        login_status: GetCookieStatus = GetCookieStatus(success=False)
        phone_number = PhoneForm.input.value
        captcha = int(self.input.value) if self.input.value.isdigit() else self.input.value

        # 1. 通过短信验证码获取 login_ticket / 使用已有 login_ticket
        if captcha:
            login_status, cookies = await get_login_ticket_by_captcha(phone_number,
                                                                      captcha,
                                                                      PhoneForm.device_id)
            if login_status:
                logger.info(f"用户 {phone_number} 成功获取 login_ticket: {cookies.login_ticket}")
                account = conf.accounts.get(cookies.bbs_uid)
                """当前的账户数据对象"""
                if not account or not account.cookies:
                    conf.accounts.update({
                        cookies.bbs_uid: UserAccount(phone_number=phone_number, cookies=cookies)
                    })
                    account = conf.accounts[cookies.bbs_uid]
                else:
                    account.cookies.update(cookies)
                fp_status, account.device_fp = await get_device_fp(account.device_id_ios)
                if fp_status:
                    logger.info(f"成功获取 device_fp: {account.device_fp}")
                conf.save()
                CaptchaLoginInformation.radio_tuple.login_ticket_by_captcha.turn_on()
        else:
            account_list = list(filter(lambda x: x.phone_number == phone_number, conf.accounts.values()))
            account = account_list[0] if account_list else None
            if not account:
                self.app.notice(f"手机号为 [bold red]{phone_number}[/] 的账户暂未被绑定！")
                self.loading.hide()
                self.button.error.show()
                self.close_login()
                return

        # 2. 通过 login_ticket 获取 stoken 和 ltoken
        if login_status or account:
            login_status, cookies = await get_multi_token_by_login_ticket(account.cookies)
            if login_status:
                logger.info(f"用户 {phone_number} 成功获取 stoken: {cookies.stoken}")
                account.cookies.update(cookies)
                conf.save()
                CaptchaLoginInformation.radio_tuple.multi_token_by_login_ticket.turn_on()

                # 3. 通过 stoken_v1 获取 stoken_v2 和 mid
                login_status, cookies = await get_stoken_v2_by_v1(account.cookies, account.device_id_ios)
                if login_status:
                    logger.info(f"用户 {phone_number} 成功获取 stoken_v2: {cookies.stoken_v2}")
                    account.cookies.update(cookies)
                    conf.save()
                    CaptchaLoginInformation.radio_tuple.get_stoken_v2.turn_on()

                    # 4. 通过 stoken_v2 获取 ltoken
                    login_status, cookies = await get_ltoken_by_stoken(account.cookies, account.device_id_ios)
                    if login_status:
                        logger.info(f"用户 {phone_number} 成功获取 ltoken: {cookies.ltoken}")
                        account.cookies.update(cookies)
                        conf.save()
                        CaptchaLoginInformation.radio_tuple.get_ltoken_by_stoken.turn_on()

                        # 5. 通过 stoken_v2 获取 cookie_token
                        login_status, cookies = await get_cookie_token_by_stoken(account.cookies, account.device_id_ios)
                        if login_status:
                            logger.info(f"用户 {phone_number} 成功获取 cookie_token: {cookies.cookie_token}")
                            account.cookies.update(cookies)
                            conf.save()
                            CaptchaLoginInformation.radio_tuple.cookie_token_by_stoken.turn_on()

                            # TODO 2023/04/12 此处如果可以模拟App的登录操作，再标记为登录完成，更安全
                            CaptchaLoginInformation.radio_tuple.login_finished.turn_on()
                            self.app.notice(f"用户 [bold green]{phone_number}[/] 登录成功！")
                            self.button.success.show()

                            ExchangePlanView.account_content.update_data()

        self.loading.hide()
        if not login_status:
            notice_text = "登录失败：[bold red]"
            if login_status.incorrect_captcha:
                notice_text += "验证码错误！"
            elif login_status.login_expired:
                notice_text += "登录失效！"
            elif login_status.incorrect_return:
                notice_text += "服务器返回错误！"
            elif login_status.network_error:
                notice_text += "网络连接失败！"
            elif login_status.missing_bbs_uid:
                notice_text += "Cookies缺少 bbs_uid（例如 ltuid, stuid）"
            elif login_status.missing_login_ticket:
                notice_text += "Cookies缺少 login_ticket！"
            elif login_status.missing_cookie_token:
                notice_text += "Cookies缺少 cookie_token！"
            elif login_status.missing_stoken:
                notice_text += "Cookies缺少 stoken！"
            elif login_status.missing_stoken_v1:
                notice_text += "Cookies缺少 stoken_v1"
            elif login_status.missing_stoken_v2:
                notice_text += "Cookies缺少 stoken_v2"
            elif login_status.missing_mid:
                notice_text += "Cookies缺少 mid"
            else:
                notice_text += "未知错误！"
            notice_text += "[/] 如果部分步骤成功，你仍然可以尝试获取收货地址、兑换等功能"
            self.button.error.show()
            self.app.notice(notice_text)

        self.close_login()
        return login_status

    async def on_input_submitted(self, _: Input.Submitted) -> None:
        await self.login()

    async def _on_button_pressed(self, event: ControllableButton.Pressed) -> None:
        if event.button.id == "login":
            # 按下“登录”按钮时触发的事件

            await self.login()

        elif event.button.id in ["login_error", "login_success"]:
            # 按下“完成（成功）”或“返回（出错）”按钮时触发的事件

            if event.button.id == "login_success":
                [i.turn_off() for i in CaptchaLoginInformation.radio_tuple]
            self.button.login.show()
            self.button.error.hide()
            self.button.success.hide()
            self.before_login = True



from __future__ import annotations

from itertools import zip_longest
from typing import Optional, Tuple

from rich.console import RenderableType
from rich.text import TextType
from textual import events
from textual.app import ComposeResult
from textual.events import Event
from textual.widgets import (
    Button,
    LoadingIndicator, RadioButton, TabbedContent, TabPane, ContentSwitcher, Tabs, ListItem
)
from textual.widgets._button import ButtonVariant
from textual.widgets._tabbed_content import ContentTab

from mys_goods_tool.custom_css import *
from mys_goods_tool.user_data import ExchangePlan


class RadioStatus(RadioButton, can_focus=False):
    """
    完成的进度节点，不允许点击交互
    可通过触发事件以即时修改value属性
    """

    class ChangeStatus(Event):
        """
        请求按钮状态修改的事件
        """

        def __init__(self, radio_status: RadioStatus):
            self.radio_status = radio_status
            super().__init__()

    class TurnOn(ChangeStatus):
        """请求按钮状态修改为亮起的事件"""
        pass

    class TurnOff(ChangeStatus):
        """请求按钮状态修改为熄灭的事件"""
        pass

    def turn_on(self):
        """修改按钮状态为亮起"""
        self.post_message(RadioStatus.TurnOn(self))

    def turn_off(self):
        """修改按钮状态为熄灭"""
        self.post_message(RadioStatus.TurnOff(self))

    def toggle(self) -> None:
        """
        重写按钮交互，交互时不会改变按钮状态
        """
        pass


class StaticStatus(Static):
    """
    实时文本说明，可通过触发事件以即时修改文本属性
    """

    class ChangeRenderable(Event):
        """
        请求renderable属性（此处与文本相关）修改的事件
        """

        def __init__(self, static_status: StaticStatus, renderable: RenderableType, text_align: Optional[str] = None):
            self.static_status = static_status
            self.renderable = renderable
            self.text_align = text_align
            super().__init__()

    def change_text(self, renderable: RenderableType, text_align: Optional[str] = None) -> None:
        """
        修改renderable属性
        主要是为了修改text_align属性，如果只是修改文本内容，直接使用update方法即可

        :param renderable: 新的 renderable 文本内容
        :param text_align: 新的 text_align 属性
        """
        self.post_message(StaticStatus.ChangeRenderable(self, renderable, text_align))


class ControllableButton(Button):
    """
    带隐藏、显示、禁用、启用控制方法的按钮
    """

    def show(self):
        """
        显示
        """
        self.display = BLOCK

    def hide(self):
        """
        隐藏
        """
        self.display = NONE

    def disable(self):
        """
        禁用
        """
        button = self
        button.disabled = True

    def enable(self):
        """
        启用
        """
        button = self
        button.disabled = False


class LoadingDisplay(LoadingIndicator):
    def show(self):
        """
        显示
        """
        self.display = BLOCK

    def hide(self):
        """
        隐藏
        """
        self.display = NONE


class DynamicTabbedContent(TabbedContent):
    """
    可动态增加TabPane的TabbedContent
    """

    def __init__(self, *titles: TextType, initial: str = "") -> None:
        super().__init__(*titles, initial=initial)
        self._tabs: Optional[Tabs] = None
        self._content_switcher = ContentSwitcher(initial=self._initial or None)

    @classmethod
    def _set_id(cls, content: TabPane, new_id: str) -> TabPane:
        """Set an id on the content, if not already present.

        Args:
            content: a TabPane.
            new_id: New `is` attribute, if it is not already set.

        Returns:
            The same TabPane.
        """
        if content.id is None:
            content.id = new_id
        return content

    def compose(self) -> ComposeResult:
        """Compose the tabbed content."""

        # Wrap content in a `TabPane` if required.
        pane_content = [
            (
                self._set_id(content, f"tab-{index}")
                if isinstance(content, TabPane)
                else TabPane(
                    title or self.render_str(f"Tab {index}"), content, id=f"tab-{index}"
                )
            )
            for index, (title, content) in enumerate(
                zip_longest(self.titles, self._tab_content), 1
            )
        ]
        # Get a tab for each pane
        tabs = [
            ContentTab(content._title, content.id or "") for content in pane_content
        ]
        # Yield the tabs
        self._tabs = Tabs(*tabs, active=self._initial or None)
        yield self._tabs
        # Yield the content switcher and panes
        with self._content_switcher:
            yield from pane_content

    async def append(self, content: TabPane):
        """Add a new TabPane to the end of the tab list.

        Args:
            content: The new TabPane object to add.
        """
        self.titles.append(content._title)
        self._tab_content.append(content)
        tab_pane_with_id = self._set_id(content, f"tab-{len(self._content_switcher.children) + 1}")
        content_tab = ContentTab(tab_pane_with_id._title, tab_pane_with_id.id or "")
        await self._content_switcher.mount(tab_pane_with_id)
        # Incase _initial is not set (empty TabbedContent), set it to the active tab
        self._content_switcher._initial = self.active
        initial = self._content_switcher._initial
        with self._content_switcher.app.batch_update():
            for child in self._content_switcher.children:
                child.display = bool(initial) and child.id == initial
        self._tabs.add_tab(content_tab)


class GameButton(ControllableButton):
    """
    商品分区视图下的商品选择按钮
    """

    def __init__(
            self,
            label: TextType | None = None,
            variant: ButtonVariant = "default",
            *,
            name: str | None = None,
            id: str | None = None,
            classes: str | None = None,
            disabled: bool = False,
            partition: Tuple[str, str]
    ):
        super().__init__(label, variant, name=name, id=id, classes=classes, disabled=disabled)
        self.partition = partition

    class Pressed(Button.Pressed):
        def __init__(self, button: GameButton):
            super().__init__(button)
            self.button = button


class PlanButton(ControllableButton):
    """
    编辑兑换计划视图下的按钮
    """

    def __init__(
            self,
            label: TextType | None = None,
            variant: ButtonVariant = "default",
            *,
            name: str | None = None,
            id: str | None = None,
            classes: str | None = None,
            disabled: bool = False,
            plan: ExchangePlan
    ):
        super().__init__(label, variant, name=name, id=id, classes=classes, disabled=disabled)
        self.plan = plan

    class Pressed(Button.Pressed):
        def __init__(self, button: PlanButton):
            super().__init__(button)
            self.button = button


class UnClickableItem(ListItem):
    """
    无法点击、不会高亮的列表项
    """

    async def _on_click(self, _: events.Click) -> None:
        pass

    def watch_highlighted(self, value: bool) -> None:
        pass



import hashlib
import json
import os
import random
import string
import time
import uuid
from multiprocessing import Pool, pool
from socket import socket, AF_INET, SOCK_STREAM
from typing import Literal, Union, Dict, List, Any, Callable, Iterable, Optional
from urllib.parse import urlencode

import httpx
import ntplib
import tenacity
from loguru import logger
from pydantic import ValidationError

from mys_goods_tool.user_data import config as conf

LOG_FORMAT: str = (
    "<g>{time:MM-DD HH:mm:ss}</g> "
    "[<lvl>{level}</lvl>] "
    "<c><u>{name}</u></c> | "
    "{message}"
)
"""默认日志格式"""

logger.remove()
if conf.preference.log_path:
    logger.add(conf.preference.log_path, diagnose=True, format=LOG_FORMAT, level="DEBUG")


def custom_attempt_times(retry: bool):
    """
    自定义的重试机制停止条件\n
    根据是否要重试的bool值，给出相应的`tenacity.stop_after_attempt`对象

    :param retry: True - 重试次数达到偏好设置中 max_retry_times 时停止; False - 执行次数达到1时停止，即不进行重试
    """
    if retry:
        return tenacity.stop_after_attempt(
            conf.preference.max_retry_times + 1) if conf.preference.max_retry_times else tenacity.stop_after_attempt(1)
    else:
        return tenacity.stop_after_attempt(1)


def get_async_retry(retry: bool):
    """
    获取异步重试装饰器

    :param retry: True - 重试次数达到偏好设置中 max_retry_times 时停止; False - 执行次数达到1时停止，即不进行重试
    """
    return tenacity.AsyncRetrying(
        stop=custom_attempt_times(retry),
        retry=tenacity.retry_if_exception_type(BaseException),
        wait=tenacity.wait_fixed(conf.preference.retry_interval),
    )


class NtpTime:
    """
    NTP时间校准相关
    """
    time_offset = 0
    """本地时间与互联网时间的偏差"""

    @classmethod
    def sync(cls):
        """
        校准时间
        """
        if conf.preference.enable_ntp_sync:
            if not conf.preference.ntp_server:
                logger.error("开启了互联网时间校对，但未配置NTP服务器 preference.ntp_server，放弃时间同步")
                return False
            try:
                for attempt in get_async_retry(True):
                    with attempt:
                        cls.time_offset = ntplib.NTPClient().request(
                            conf.preference.ntp_server).tx_time - time.time()
            except tenacity.RetryError:
                logger.exception("校对互联网时间失败，改为使用本地时间")
                return False
            logger.info("互联网时间校对完成")
            return True
        else:
            logger.info("未开启互联网时间校对，跳过时间同步")
            return True

    @classmethod
    def time(cls) -> float:
        """
        获取校准后的时间（如果校准成功）
        """
        return time.time() + cls.time_offset


def generate_device_id() -> str:
    """
    生成随机的x-rpc-device_id
    """
    return str(uuid.uuid4()).upper()


def cookie_str_to_dict(cookie_str: str) -> Dict[str, str]:
    """
    将字符串Cookie转换为字典Cookie
    """
    cookie_str = cookie_str.replace(" ", "")
    # Cookie末尾缺少 ; 的情况
    if cookie_str[-1] != ";":
        cookie_str += ";"

    cookie_dict = {}
    start = 0
    while start != len(cookie_str):
        mid = cookie_str.find("=", start)
        end = cookie_str.find(";", mid)
        cookie_dict.setdefault(cookie_str[start:mid], cookie_str[mid + 1:end])
        start = end + 1
    return cookie_dict


def cookie_dict_to_str(cookie_dict: Dict[str, str]) -> str:
    """
    将字符串Cookie转换为字典Cookie
    """
    cookie_str = ""
    for key in cookie_dict:
        cookie_str += (key + "=" + cookie_dict[key] + ";")
    return cookie_str


def generate_ds(data: Union[str, dict, list, None] = None, params: Union[str, dict, None] = None,
                platform: Literal["ios", "android"] = "ios", salt: Optional[str] = None):
    """
    获取Headers中所需DS

    :param data: 可选，网络请求中需要发送的数据
    :param params: 可选，URL参数
    :param platform: 可选，平台，ios或android
    :param salt: 可选，自定义salt
    """
    if data is None and params is None or \
            salt is not None and salt != conf.salt_config.SALT_PROD:
        if platform == "ios":
            salt = salt or conf.salt_config.SALT_IOS
        else:
            salt = salt or conf.salt_config.SALT_ANDROID
        t = str(int(NtpTime.time()))
        a = "".join(random.sample(
            string.ascii_lowercase + string.digits, 6))
        re = hashlib.md5(
            f"salt={salt}&t={t}&r={a}".encode()).hexdigest()
        return f"{t},{a},{re}"
    else:
        if params:
            salt = conf.salt_config.SALT_PARAMS if not salt else salt
        else:
            salt = conf.salt_config.SALT_DATA if not salt else salt

        if not data:
            if salt == conf.salt_config.SALT_PROD:
                data = {}
            else:
                data = ""
        if not params:
            params = ""

        if not isinstance(data, str):
            data = json.dumps(data)
        if not isinstance(params, str):
            params = urlencode(params)

        t = str(int(time.time()))
        r = str(random.randint(100000, 200000))
        c = hashlib.md5(
            f"salt={salt}&t={t}&r={r}&b={data}&q={params}".encode()).hexdigest()
        return f"{t},{r},{c}"


def generate_seed_id(length: int = 8) -> str:
    """
    生成随机的 seed_id（即长度为8的十六进制数）

    :param length: 16进制数长度
    """
    max_num = int("FF" * length, 16)
    return hex(random.randint(0, max_num))[2:]


def generate_fp_locally(length: int = 13):
    """
    于本地生成 device_fp

    :param length: device_fp 长度
    """
    characters = string.digits + "abcdef"
    return ''.join(random.choices(characters, k=length))


async def get_file(url: str, retry: bool = True):
    """
    下载文件

    :param url: 文件URL
    :param retry: 是否允许重试
    :return: 文件数据
    """
    try:
        async for attempt in get_async_retry(retry):
            with attempt:
                async with httpx.AsyncClient() as client:
                    res = await client.get(url, timeout=conf.preference.timeout, follow_redirects=True)
                return res.content
    except tenacity.RetryError:
        logger.exception(f"下载文件 - {url} 失败")


PORT_RANGE = (1024, 49151)
"""
随机选择的端口所在范围

端口范围说明：https://www.rfc-editor.org/rfc/rfc6335.html#section-6
为何选择这个端口范围：https://www.rfc-editor.org/rfc/rfc6335.html#section-8.1.2
"""


def get_free_port(host: str = "localhost"):
    """
    在范围内随机获取一个空闲端口
    """
    used_port = set()
    while len(used_port) != PORT_RANGE[1] - PORT_RANGE[0] + 1:
        port = random.randint(PORT_RANGE[0], PORT_RANGE[1])
        with socket(AF_INET, SOCK_STREAM) as sock:
            res = sock.connect_ex((host, port))
            if res:
                return port
            else:
                used_port.add(port)
    logger.error("获取随机可用端口 - 未找到可用端口")
    return


class Subscribe:
    """
    在线配置相关(需实例化)
    """
    FILE_URL = "https://github.com/Ljzd-PRO/Mys_Goods_Tool/raw/dev/subscribe/configs.json"
    CONFIG_URL = os.path.join(conf.preference.github_proxy, FILE_URL) if conf.preference.github_proxy else FILE_URL
    conf_list: List[Dict[str, Any]] = []
    '''当前插件版本可用的配置资源'''

    def __init__(self):
        self.index = 0

    @classmethod
    async def download(cls) -> bool:
        """
        读取在线配置资源
        :return: 是否成功
        """
        try:
            for attempt in get_async_retry(True):
                with attempt:
                    file = await get_file(cls.CONFIG_URL)
                    file = json.loads(file.decode())
                    if not file:
                        return False
                    cls.conf_list = list(
                        filter(lambda co: ..., file))
                    cls.conf_list.sort(
                        key=lambda co: conf["time"], reverse=True)
                    return True
        except (json.JSONDecodeError, KeyError):
            logger.error(f"获取在线配置资源 - 解析文件失败")
            return False

    async def load(self, force: bool = False) -> bool:
        """
        优先加载来自网络的配置，若获取失败，则返回本地默认配置。\n
        若下载失败返回`False`

        :param force: 是否强制在线读取配置，而不使用本地缓存的
        """
        success = True
        if not Subscribe.conf_list or force or self.index >= len(Subscribe.conf_list):
            logger.info(f"读取配置 - 开始下载配置...")
            success = await self.download()
            self.index = 0
        if not success:
            return False
        else:
            try:
                conf.device_config.parse_obj(Subscribe.conf_list[self.index]["config"])
            except ValidationError:
                logger.exception(f"获取在线配置资源 - 加载配置失败")
                return False

            self.index += 1
            return True


class ProcessManager:
    """
    异步进程管理器
    """

    def __init__(self, callback: Optional[Callable] = None, error_callback: Optional[Callable] = None):
        """
        创建进程池，初始化异步进程管理器，包含进程池对象

        :param callback: 线程任务正常执行结束后的回调函数
        :param error_callback: 线程任务执行发生错误后的回调函数
        """
        self.pool: Optional[pool.Pool] = None
        """进程池"""
        self.callback = callback
        """线程任务正常执行结束后的回调函数"""
        self.error_callback = error_callback
        """线程任务执行发生错误后的回调函数"""

    def start(self, process_func: Callable, process_params: Iterable):
        """
        并启动线程任务
        """
        self.pool = Pool(1)
        self.pool.apply_async(process_func, process_params, callback=self.callback, error_callback=self.error_callback)
        self.pool.close()



from __future__ import annotations

from textual.containers import Container, Horizontal, Vertical
from textual.dom import DOMNode
from textual.screen import Screen
from textual.widgets import (
    DataTable,
    Static,
    TextLog, Tree
)

NONE = "none"
"""隐藏（CSS display 属性 - none）"""
BLOCK = "block"
"""显示（CSS display 属性 - block）"""

DOMNode.DEFAULT_CSS += """
    * {
        transition: background 500ms in_out_cubic, color 500ms in_out_cubic;
    }
    """
Screen.DEFAULT_CSS += """
    Screen {
        layers: base overlay notes notifications;
        overflow: hidden;
    }

    Screen>Container {
        height: 100%;
        overflow: hidden;
    }
    """
TextLog.DEFAULT_CSS += """
    TextLog {
        background: $surface;
        color: $text;
        height: 50vh;
        dock: bottom;
        layer: notes;
        border-top: hkey $primary;
        offset-y: 0;
        transition: offset 400ms in_out_cubic;
        padding: 0 1 1 1;
    }

    TextLog:focus {
        offset: 0 0 !important;
    }

    TextLog.-hidden {
        offset-y: 100%;
    }
    """
DataTable.DEFAULT_CSS += """
    DataTable {
        height: 16;
    }
    """
Tree.DEFAULT_CSS += """
    Tree {
        margin: 1 0;
    }
    """
Vertical.DEFAULT_CSS += """
    Vertical {
        height: auto;
        layout: vertical;
    }
    """
Horizontal.DEFAULT_CSS += """
    Horizontal {
        height: auto;
        layout: horizontal;
    }
    """


class Body(Container):
    DEFAULT_CSS = """
    Body {
        height: 100%;
        overflow-y: scroll;
        width: 100%;
        background: $surface;
    }
    """


class Title(Static):
    DEFAULT_CSS = """
    Sidebar Title {
        background: $boost;
        color: $secondary;
        padding: 2 4;
        border-right: vkey $background;
        dock: top;
        text-align: center;
        text-style: bold;
    }
    """


class SectionTitle(Static):
    DEFAULT_CSS = """
    SectionTitle {
        padding: 1 2;
        background: $boost;
        text-align: center;
        text-style: bold;
    }
    """


class Message(Static):
    DEFAULT_CSS = """
    Message {
        margin: 0 1;
    }
    """


class AboveFold(Container):
    DEFAULT_CSS = """
    AboveFold {
        width: 100%;
        height: auto;
        align: center middle;
    }
    """


class Section(Container):
    DEFAULT_CSS = """
    Section {
        height: auto;
        min-width: 40;
        margin: 1 2 4 2;
    }
    """


class Column(Container):
    DEFAULT_CSS = """
    Column {
        height: auto;
        min-height: 100vh;
        align: center top;
        overflow: hidden;
    }
    """


class TextContent(Static):
    DEFAULT_CSS = """
    TextContent {
        margin: 1 0;
    }
    """


class QuickAccess(Container):
    DEFAULT_CSS = """
        QuickAccess {
        width: 30;
        dock: left;
    }
    """


class Window(Container):
    DEFAULT_CSS = """
    Window {
        background: $boost;
        overflow: auto;
        height: auto;
        max-height: 16;
    }

    Window>Static {
        width: auto;
    }
    """


class SubTitle(Static):
    DEFAULT_CSS = """
        SubTitle {
        padding-top: 1;
        border-bottom: heavy $panel;
        color: $text;
        text-style: bold;
    }
    """


class LoginForm(Container):
    """
    登录表单
    """
    DEFAULT_CSS = """
    LoginForm {
        height: auto;
        padding: 1 2;
        layout: grid;
        grid-size: 2;
        grid-rows: 4;
        grid-columns: 12 1fr;
        background: $boost;
        border: wide $background;
    }

    LoginForm ControllableButton {
        margin: 0 1;
        width: 100%;
    }

    LoginForm .label {
        padding: 1 2;
        text-align: right;
    }
    """


class ExchangePlanContent(Container):
    """
    兑换计划的添加步骤视图
    """
    DEFAULT_CSS = """
    ExchangePlanContent {
        height: auto;
        width: 1fr;
        border: round #666;
        padding: 2;
    }

    ExchangePlanContent OptionList {
        height: auto;
        width: 1fr;
        padding: 1;
        border: round $primary;
    }
    
    ExchangePlanContent OptionList Option {
        margin: 1;
    }
    
    ExchangePlanContent ListView {
        height: auto;
        width: 1fr;
        padding: 1;
        border: round $primary;
    }
    
    ExchangePlanContent StaticStatus {
        margin: 0 0 1 0;
    }
    
    ExchangePlanContent Horizontal Button {
        margin: 0 1 0 0;
    }
    
    ExchangePlanContent Horizontal LoadingIndicator {
        width: auto;
        min-width: 16;
        height: 3;
    }
    """


class CaptchaTips(Container):
    """
    登陆信息面板文本视图
    """
    DEFAULT_CSS = """
    CaptchaTips {
        height: 100%;
        width: 1fr;
        align: right middle;
        padding: 1;
        overflow: auto;
        border: round #666;
    }

    App.-light-mode Tips {
        border: round #CCC;
    }

    CaptchaTips StaticStatus {
        width: 100%;
        align: center top;
        text-align: center;
    }
    """


class CaptchaStepSet(Container):
    """
    登陆进度节点集合视图
    """
    DEFAULT_CSS = """
    CaptchaStepSet {
        height: auto;
        width: 1fr;
        align: left middle;
        overflow: auto;
        border: round #666;
    }

    App.-light-mode StepSet {
        border: round #CCC;
    }

    CaptchaStepSet RadioStatus {
        margin: 1 1;
    }
    """



from __future__ import annotations

import sys
from io import StringIO

from rich.console import RenderableType
from rich.markdown import Markdown
from rich.text import Text
from textual import events
from textual.app import App, ComposeResult, DEFAULT_COLORS
from textual.binding import Binding
from textual.color import Color
from textual.events import Event
from textual.reactive import reactive
from textual.widgets import (
    Button,
    Footer,
    Header,
    Switch
)

from mys_goods_tool.custom_css import *
from mys_goods_tool.custom_widget import RadioStatus, StaticStatus
from mys_goods_tool.exchange_mode import ExchangeModeView, EnterExchangeMode, ExitExchangeMode
from mys_goods_tool.exchange_plan_view import ExchangePlanView
from mys_goods_tool.login_view import LoginView
from mys_goods_tool.user_data import ROOT_PATH, VERSION, different_device_and_salt
from mys_goods_tool.utils import LOG_FORMAT, logger

WELCOME_MD = """
# Mys_Goods_Tool - 米游社商品兑换工具

## 更新说明

- 兑换请求Headers增加与修改了 `Referer`, `x-rpc-device_fp`, `x-rpc-verify_key`, `Origin` 等字段，可能修复兑换失败的问题
- 修复登陆时因为连接断开（client has been closed）而导致登陆失败的问题
- 防止因配置文件中默认存在 `device_config`, `salt_config` 而导致更新后默认配置被原配置覆盖的问题
- 若需要修改 `device_config` 配置，修改后还设置用户数据文件中 `preference.override_device_and_salt` 为 `true` 以覆盖默认值
- 修复Unix下即使安装了 uvloop 也找不到，无法应用的问题

## 功能和特性

- 使用 [Textual](https://github.com/Textualize/textual) 终端图形界面库，支持 Windows / Linux / macOS 甚至可能是移动端SSH客户端
- 短信验证码登录（只需接收一次验证码）
- 内置人机验证页面，无需前往官网验证
- 多账号支持
- 支持米游社所有分区的商品兑换

## 其他
- [**🔗完整说明文档**](https://github.com/Ljzd-PRO/Mys_Goods_Tool/wiki)
- 仅供学习时参考

- 相似项目推荐:  \
  **mysTool - 米游社辅助工具插件**  \
  简介：NoneBot2 插件 | 米游社工具-每日米游币任务、游戏签到、商品兑换、免抓包登录、原神树脂提醒  \
  🔗 https://github.com/Ljzd-PRO/nonebot-plugin-mystool
  
- 本项目已开启[🔗Github Actions](https://github.com/Ljzd-PRO/Mys_Goods_Tool/actions)。
  欢迎[🔗指出Bug](https://github.com/Ljzd-PRO/Mys_Goods_Tool/issues)
  和[🔗贡献代码](https://github.com/Ljzd-PRO/Mys_Goods_Tool/pulls)👏

- 开发版分支：[🔗dev](https://github.com/Ljzd-PRO/Mys_Goods_Tool/tree/dev/)
"""


class Welcome(Container):
    DEFAULT_CSS = """
    Welcome {
        background: $boost;
        height: auto;
        max-width: 100;
        min-width: 40;
        border: wide $primary;
        padding: 1 2;
        margin: 1 2;
        box-sizing: border-box;
    }

    Welcome Button {
        width: 100%;
        margin-top: 1;
    }
    """

    def compose(self) -> ComposeResult:
        yield Static(Markdown(WELCOME_MD))
        yield Button("开始使用", variant="success")

    def _on_button_pressed(self) -> None:
        self.app.query_one(".location-first").scroll_visible(duration=0.5, top=True)


class Version(Static):
    DEFAULT_CSS = """
    Version {
        color: $text-disabled;
        dock: bottom;
        text-align: center;
        padding: 1;
    }
    """

    def render(self) -> RenderableType:
        return f"[b]v{VERSION}"


class Sidebar(Container):
    DEFAULT_CSS = """
    Sidebar {
        width: 40;
        background: $panel;
        transition: offset 500ms in_out_cubic;
        layer: overlay;
    }

    Sidebar:focus-within {
        offset: 0 0 !important;
    }

    Sidebar.-hidden {
        offset-x: -100%;
    }

    Sidebar Title {
        background: $boost;
        color: $secondary;
        padding: 2 4;
        border-right: vkey $background;
        dock: top;
        text-align: center;
        text-style: bold;
    }
    """

    def compose(self) -> ComposeResult:
        yield Title("Mys_Goods_Tool")
        with Container():
            yield Message("米游社商品兑换工具")
            yield Message()
            yield Message(
                "[bold italic black][link=https://github.com/Ljzd-PRO/Mys_Goods_Tool]🔗 GitHub 项目链接[/link][/]")
            yield Message()
            yield Message(
                "[bold italic black][link=https://github.com/Ljzd-PRO/nonebot-plugin-mystool]🔗 mysTool - 米游社辅助工具 NoneBot机器人插件[/link][/]")
            yield Version()
        yield DarkSwitch()


class LocationLink(Static):
    DEFAULT_CSS = """
    LocationLink {
        margin: 1 0 0 1;
        height: 1;
        padding: 1 2;
        background: $boost;
        color: $text;
        box-sizing: content-box;
        content-align: center middle;
    }

    LocationLink:hover {
        background: $accent;
        color: $text;
        text-style: bold;
    }

    .pad {
        margin: 1 0;
    }
    """

    def __init__(self, label: str, reveal: str) -> None:
        super().__init__(label)
        self.reveal = reveal

    def _on_click(self, _: events.Click) -> None:
        self.app.query_one(self.reveal).scroll_visible(top=True, duration=0.5)


class DarkSwitch(Horizontal):
    DEFAULT_CSS = """
    DarkSwitch {
        background: $panel;
        padding: 1;
        dock: bottom;
        height: auto;
        border-right: vkey $background;
    }

    DarkSwitch .label {
        width: 1fr;
        padding: 1 2;
        color: $text-muted;
    }

    DarkSwitch Switch {
        background: $boost;
        dock: left;
    }
    """

    # TODO: 目前似乎切换后会导致UI界面卡顿，待优化

    def compose(self) -> ComposeResult:
        yield Switch(value=self.app.dark)
        yield Static("深色模式切换", classes="label")

    def _on_mount(self, _: events.Mount) -> None:
        self.watch(self.app, "dark", self.on_dark_change, init=False)

    def on_dark_change(self) -> None:
        self.query_one(Switch).value = self.app.dark

    def on_switch_changed(self, event: Switch.Changed) -> None:
        self.app.dark = event.value


class Notification(Static):
    """
    通知消息框
    """
    DEFAULT_CSS = """
    Notification {
        dock: bottom;
        layer: notification;
        width: auto;
        margin: 2 4;
        padding: 1 2;
        background: $background;
        color: $text;
        height: auto;
        border: wide $primary;
    }
    """

    def _on_mount(self, _: events.Mount) -> None:
        self.set_timer(3, self.remove)

    def on_click(self) -> None:
        self.remove()


# 主题颜色
# https://colorhunt.co/palette/b9eddd87cbb9569daa577d86
# TODO: 希望可以找到更好的日间模式(Light Mode)配色方案
DEFAULT_COLORS["dark"].primary = Color.parse("#569DAA")
DEFAULT_COLORS["dark"].secondary = Color.parse("#577D86")
DEFAULT_COLORS["dark"].accent = DEFAULT_COLORS["dark"].primary


# DEFAULT_COLORS["light"].primary = Color.parse("#B9EDDD")
# DEFAULT_COLORS["light"].secondary = Color.parse("#87CBB9")
# DEFAULT_COLORS["light"].accent = DEFAULT_COLORS["dark"].primary


class TuiApp(App):
    TITLE = "Mys_Goods_Tool"
    """textual TUI 标题"""
    BINDINGS = [
        ("ctrl+b", "toggle_sidebar", "🧭侧栏"),
        ("ctrl+t", "app.toggle_dark", "🌓深色模式切换"),
        ("ctrl+s", "app.screenshot()", "✂截屏"),
        ("f1", "app.toggle_class('TextLog', '-hidden')", "📃日志"),
        Binding("ctrl+c,ctrl+q", "app.quit", "🚪退出", show=True),
    ]
    """按键绑定"""

    show_sidebar = reactive(False)

    app: TuiApp
    """当前App实例"""

    text_log_writer: TextLogWriter
    """textual日志输出流"""
    text_log = TextLog(classes="-hidden", wrap=False, highlight=True, markup=True)
    """textual日志输出界面"""

    quick_access = QuickAccess(
        LocationLink("🏠 主页", ".location-top"),
        LocationLink("🔑 登录绑定", ".location-login"),
        LocationLink("📅 管理兑换计划", ".location-add_plan"),
        LocationLink("⏰ 进入兑换模式", ".location-exchange_mode"),
    )
    """快速访问菜单"""
    disable_required_column = (
        AboveFold(Welcome(), classes="location-top"),
        Column(
            Section(
                SectionTitle("米游社账号登录绑定"),
                LoginView(),
            ),
            classes="location-login location-first",
        ),
        Column(
            Section(
                SectionTitle("管理米游币商品兑换计划"),
                ExchangePlanView(),
            ),
            classes="location-add_plan",
        )
    )
    """进入兑换模式后需要禁用的Column"""
    body = Body(
        quick_access,
        *disable_required_column,
        Column(
            Section(
                SectionTitle("定时兑换模式"),
                ExchangeModeView(),
            ),
            classes="location-exchange_mode",
        )
    )
    """主体内容"""

    def notice(self, renderable: RenderableType) -> None:
        """
        发出消息通知

        :param renderable: 通知内容
        """
        self.app.screen.mount(Notification(renderable))

    def add_note(self, renderable: RenderableType) -> None:
        """
        输出至日志（仅textual TUI内，而不是loguru的Logger）

        :param renderable: 日志内容
        """
        self.query_one(TextLog).write(renderable)

    def compose(self) -> ComposeResult:
        yield Container(
            Sidebar(classes="-hidden"),
            Header(show_clock=False),
            self.text_log,
            self.body
        )
        yield Footer()

    async def on_event(self, event: Event) -> None:
        """
        重写事件处理，在收到请求修改Widget属性的事件时，完成修改
        这是因为组件只会在事件结束后进行刷新，如果有事件需要修改多个组件属性，就无法一个个生效，需要交由新的事件处理。

        :param event: 事件
        """
        if isinstance(event, RadioStatus.TurnOn):
            event.radio_status.value = True
        elif isinstance(event, RadioStatus.TurnOff):
            event.radio_status.value = False
        elif isinstance(event, StaticStatus.ChangeRenderable):
            event.static_status.update(event.renderable)
            if event.text_align:
                event.static_status.styles.text_align = event.text_align

        elif isinstance(event, EnterExchangeMode):
            self.quick_access.disabled = True
            for column in self.disable_required_column:
                column.disabled = True
                column.display = NONE
            self.app.query_one(".location-exchange_mode").scroll_visible(top=True, duration=0.5, force=True)

        elif isinstance(event, ExitExchangeMode):
            self.quick_access.disabled = False
            for column in self.disable_required_column:
                column.disabled = False
                column.display = BLOCK
            self.app.query_one(".location-exchange_mode").scroll_visible(top=True, animate=False)

        await super().on_event(event)

    def action_open_link(self, link: str) -> None:
        """
        跳转浏览器打开URL链接
        """
        self.app.bell()
        import webbrowser

        webbrowser.open(link)

    def action_toggle_sidebar(self) -> None:
        """
        切换侧栏
        """
        sidebar = self.query_one(Sidebar)
        self.set_focus(None)
        if sidebar.has_class("-hidden"):
            sidebar.remove_class("-hidden")
        else:
            if sidebar.query("*:focus"):
                self.screen.set_focus(None)
            sidebar.add_class("-hidden")

    class TextLogWriter(StringIO):
        def write(self, text: str) -> None:
            super().write(text)
            TuiApp.text_log.write(text)

    def _on_mount(self, _: events.Mount) -> None:
        TuiApp.app = self
        TuiApp.text_log_writer = TuiApp.TextLogWriter()
        logger.add(self.text_log_writer, diagnose=False, level="DEBUG", format=LOG_FORMAT)
        if sys.platform not in ('win32', 'cygwin', 'cli'):
            try:
                import uvloop
            except ModuleNotFoundError:
                logger.info("在非 Windows 环境下，你可以安装 uvloop 以提高性能")
            else:
                import asyncio
                asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        self.query_one("Welcome Button", Button).focus()
        if different_device_and_salt:
            logger.warning("检测到设备信息配置 device_config 或 salt_config 使用了非默认值，"
                           "如果你修改过这些配置，需要设置 preference.override_device_and_salt 为 True 以覆盖默认值并生效。"
                           "如果继续，将可能保存默认值到配置文件。")

    def action_screenshot(self, filename: str | None = None, path: str = str(ROOT_PATH)) -> None:
        """Save an SVG "screenshot". This action will save an SVG file containing the current contents of the screen.

        Args:
            filename: Filename of screenshot, or None to auto-generate. Defaults to None.
            path: Path to directory. Defaults to "./".
        """
        self.bell()
        path = self.save_screenshot(filename, path)
        message = Text.assemble("截屏已保存至 ", (f"'{path}'", "bold green"))
        self.add_note(message)
        self.screen.mount(Notification(message))



import random
import sys
import threading
import time
from datetime import datetime
from typing import Optional, Union, Tuple, Dict, List
from urllib.parse import urlparse

import ping3
from apscheduler.events import JobExecutionEvent, EVENT_JOB_EXECUTED
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.schedulers.base import BaseScheduler
from apscheduler.schedulers.blocking import BlockingScheduler
from rich.console import RenderableType
from textual import events
from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.events import Event
from textual.reactive import reactive
from textual.widgets import Static, ListView, ListItem

from mys_goods_tool.api import URL_EXCHANGE, good_exchange_sync
from mys_goods_tool.custom_widget import ControllableButton, UnClickableItem
from mys_goods_tool.data_model import ExchangeStatus
from mys_goods_tool.user_data import config as conf, ExchangePlan, Preference, ExchangeResult, different_device_and_salt
from mys_goods_tool.utils import logger, LOG_FORMAT


# TODO: ntp 时间同步

def _get_api_host() -> Optional[str]:
    """
    获取商品兑换API服务器地址
    """
    hostname = urlparse(URL_EXCHANGE).hostname
    return str(hostname) if hostname else None


def _connection_test():
    """
    连接测试
    """
    hostname = _get_api_host()
    if not hostname:
        logger.warning("商品兑换API服务器地址解析失败")
        return False
    result = ping3.ping(hostname, unit="ms")
    if result is None:
        logger.info(f"Ping 商品兑换API服务器 {hostname} 超时")
    elif result is False:
        logger.info(f"Ping 商品兑换API服务器 {hostname} 失败")
    return result


def set_scheduler(scheduler: BaseScheduler):
    """
    向兑换计划调度器添加兑换任务以及ping循环

    :param scheduler 要修改的调度器，调度器对象自身将被修改
    :return 传入的scheduler的引用
    """
    scheduler.configure(timezone=conf.preference.timezone or Preference.timezone)

    if conf.preference.enable_connection_test and scheduler.get_job("exchange-connection_test") is None:
        interval = conf.preference.connection_test_interval or Preference.connection_test_interval
        scheduler.add_job(_connection_test, "interval", seconds=interval, id=f"exchange-connection_test")

    existed_job = scheduler.get_jobs()
    for plan in conf.exchange_plans:
        job_id_start = f"exchange-plan-{plan.__hash__()}"
        # 如果已经存在相同兑换计划，就不再添加
        if not any(job.id.startswith(job_id_start) for job in existed_job):
            for i in range(1, conf.preference.exchange_thread_count + 1):
                scheduler.add_job(exchange_begin,
                                  "date",
                                  args=[plan],
                                  run_date=datetime.fromtimestamp(plan.good.time),
                                  id=f"{job_id_start}-{i}"
                                  )
            if any(job.id.startswith(job_id_start) for job in scheduler.get_jobs()):
                logger.info(f"已添加定时兑换任务 {plan.account.bbs_uid}"
                            f" - {plan.good.general_name}"
                            f" - {plan.good.time_text}")
            else:
                logger.error(f"添加兑换任务失败 {plan.account.bbs_uid}"
                             f" - {plan.good.general_name}"
                             f" - {plan.good.time_text}")

    return scheduler


def exchange_begin(plan: ExchangePlan):
    """
    到点后执行兑换

    :param plan: 兑换计划
    """
    random_x, random_y = conf.preference.exchange_latency
    latency = random.uniform(random_x, random_y)
    time.sleep(latency)
    result = good_exchange_sync(plan)
    return result


def exchange_mode_simple():
    """
    普通文本模式（无Textual）
    """
    logger.add(sys.stdout, diagnose=True, format=LOG_FORMAT, level="DEBUG")
    if not conf.exchange_plans:
        logger.info("无兑换计划需要执行")
        return

    if different_device_and_salt:
        logger.warning("检测到设备信息配置 device_config 或 salt_config 使用了非默认值，"
                       "如果你修改过这些配置，需要设置 preference.override_device_and_salt 为 True 以覆盖默认值并生效。"
                       "如果继续，将可能保存默认值到配置文件。")

    scheduler = set_scheduler(BlockingScheduler())
    finished: Dict[ExchangePlan, List[bool]] = dict(map(lambda x: (x, []), conf.exchange_plans))
    """每个兑换计划的结果"""
    lock = threading.Lock()

    @lambda func: scheduler.add_listener(func, EVENT_JOB_EXECUTED)
    def on_executed(event: JobExecutionEvent):
        """
        接收兑换结果
        """
        if event.job_id.startswith("exchange-plan"):
            thread_id = int(event.job_id.split('-')[-1])
            result: Tuple[ExchangeStatus, Optional[ExchangeResult]] = event.retval
            exchange_status, exchange_result = result

            if not exchange_status:
                hash_value = int(event.job_id.split('-')[-2])
                plan = filter(lambda x: x.__hash__() == hash_value, conf.exchange_plans)
                plan = next(plan)
                with lock:
                    finished[plan].append(False)
                    logger.error(
                        f"用户 {plan.account.bbs_uid}"
                        f" - {plan.good.general_name}"
                        f" - 线程 {thread_id}"
                        f" - 兑换请求发送失败")
                    if len(finished[plan]) == conf.preference.exchange_thread_count:
                        try:
                            conf.exchange_plans.remove(plan)
                        except KeyError:
                            pass
                        else:
                            conf.save()

            else:
                plan = exchange_result.plan
                with lock:
                    # 如果已经有一个线程兑换成功，就不再接收结果
                    if True not in finished[plan]:
                        if exchange_result.result:
                            finished[plan].append(True)
                            logger.info(
                                f"用户 {plan.account.bbs_uid}"
                                f" - {plan.good.general_name}"
                                f" - 线程 {thread_id}"
                                f" - 兑换成功")
                        else:
                            finished[plan].append(False)
                            logger.error(
                                f"用户 {plan.account.bbs_uid}"
                                f" - {plan.good.general_name}"
                                f" - 线程 {thread_id}"
                                f" - 兑换失败")

                    if len(finished[plan]) == conf.preference.exchange_thread_count:
                        try:
                            conf.exchange_plans.remove(plan)
                        except KeyError:
                            pass
                        else:
                            conf.save()

        elif event.job_id == "exchange-connection_test":
            result: Union[float, bool, None] = event.retval
            if result:
                print(
                    f"Ping 商品兑换API服务器 {_get_api_host() or 'N/A'} - 延迟 {round(result, 2) if result else 'N/A'} ms")

    try:
        logger.info("启动兑换计划定时器")
        scheduler.start()

    except KeyboardInterrupt:
        logger.info("停止兑换计划定时器")
        scheduler.shutdown()


class EnterExchangeMode(Event):
    """
    进入兑换模式的事件
    """
    pass


class ExitExchangeMode(Event):
    """
    退出兑换模式的事件
    """
    pass


class ExchangeModeWarning(Static):
    """
    进入/退出 兑换模式的提示文本
    """
    DEFAULT_CSS = """
    ExchangeModeWarning {
        width: 3fr;
    }
    """
    ENTER_TEXT = "确定要[bold]进入[/]兑换模式？进入兑换模式后[bold]无法使用其他功能[/]，定时兑换任务将会启动。你随时都可以退出，但定时任务将会暂停。"
    EXIT_TEXT = "已进入兑换模式，你可以随时[bold]退出[/]。退出后[bold]定时兑换任务将会暂停[/]。"
    display_text = reactive(ENTER_TEXT)

    def render(self) -> RenderableType:
        return self.display_text


class ExchangeModeView(Container):
    """
    兑换模式视图
    """
    DEFAULT_CSS = """
    ExchangeModeView {
        height: auto;
        width: 1fr;
        border: round #666;
        padding: 1;
        margin: 1 0;
    }
    
    ExchangeModeView ControllableButton {
        margin: 0 1;
        width: 1fr;
    }
    
    ExchangeModeView Horizontal {
        padding: 1;
        border: round #666;
    }
    
    ExchangeModeView ListView {
        overflow: hidden;
        height: auto;
    }
    """

    button_enter = ControllableButton("确定", variant="warning", id="button-exchange_mode-enter")
    button_exit = ControllableButton("退出", variant="error", id="button-exchange_mode-exit")
    button_refresh = ControllableButton("刷新", id="button-exchange_mode-refresh")
    button_exit.hide()
    warning_text = ExchangeModeWarning()
    """进入/退出 兑换模式的提示文本"""
    empty_data_item = ListItem(Static("暂无兑换计划，你可以尝试刷新"))
    list_view = ListView(empty_data_item)
    """兑换计划列表"""

    scheduler = BackgroundScheduler()
    """兑换计划调度器"""
    lock = threading.Lock()
    finished: Dict[ExchangePlan, List[bool]] = {}
    """所有的兑换结果"""

    def compose(self) -> ComposeResult:
        with Horizontal():
            yield self.warning_text
            yield self.button_enter
            yield self.button_exit
            yield self.button_refresh
        yield Static()
        yield ExchangeModePing()
        yield self.list_view

    async def update_data(self):
        """
        更新兑换计划列表
        """
        self.finished.clear()
        ExchangeResultRow.rows.clear()
        await self.list_view.clear()
        for plan in conf.exchange_plans:
            await self.list_view.append(ExchangeResultRow(plan))
            self.finished.setdefault(plan, [])
        if not conf.exchange_plans:
            await self.list_view.append(self.empty_data_item)
        set_scheduler(self.scheduler)

    @classmethod
    def on_executed(cls, event: JobExecutionEvent):
        """
        接收兑换结果
        """
        try:
            if event.job_id.startswith("exchange-plan"):
                result: Tuple[ExchangeStatus, Optional[ExchangeResult]] = event.retval
                exchange_status, exchange_result = result
                thread_id = int(event.job_id.split('-')[-1])
                if not exchange_status:
                    hash_value = int(event.job_id.split('-')[-2])
                    plan = filter(lambda x: x.__hash__() == hash_value, conf.exchange_plans)
                    plan = next(plan)
                    row = ExchangeResultRow.rows[plan]
                    with cls.lock:
                        cls.finished[plan].append(False)
                        logger.error(
                            f"用户 {plan.account.bbs_uid}"
                            f" - {plan.good.general_name}"
                            f" - 线程 {thread_id}"
                            f" - 兑换失败")
                        text = f"[bold red]💦 线程 {thread_id} - 兑换请求失败[/] "
                        row.result_preview._add_children(ExchangeResultRow.get_result_static(text))
                        row.result_preview.refresh()
                        if len(cls.finished[plan]) == conf.preference.exchange_thread_count:
                            try:
                                conf.exchange_plans.remove(plan)
                            except KeyError:
                                pass
                            else:
                                conf.save()
                else:
                    plan = exchange_result.plan
                    with cls.lock:
                        # 如果已经有一个线程兑换成功，就不再接收结果
                        if True not in cls.finished[plan]:
                            row = ExchangeResultRow.rows[plan]
                            if exchange_result.result:
                                cls.finished[plan].append(True)
                                logger.info(
                                    f"用户 {plan.account.bbs_uid}"
                                    f" - {plan.good.general_name}"
                                    f" - 线程 {thread_id}"
                                    f" - 兑换成功")
                                text = f"[bold green]🎉 线程 {thread_id} - 兑换成功[/] "
                            else:
                                cls.finished[plan].append(False)
                                logger.error(
                                    f"用户 {plan.account.bbs_uid}"
                                    f" - {plan.good.general_name}"
                                    f" - 线程 {thread_id}"
                                    f" - 兑换失败")
                                text = f"[bold red]💦 线程 {thread_id} - 兑换失败[/] "

                            row.result_preview._add_children(ExchangeResultRow.get_result_static(text))
                            row.result_preview.refresh()

                        if len(cls.finished[plan]) == conf.preference.exchange_thread_count:
                            try:
                                conf.exchange_plans.remove(plan)
                            except KeyError:
                                pass
                            else:
                                conf.save()
        except:
            logger.exception("接收兑换结果失败")

    async def _on_button_pressed(self, event: ControllableButton.Pressed):
        if event.button.id == "button-exchange_mode-enter":
            await self.update_data()
            self.button_refresh.disable()
            self.button_enter.hide()
            self.button_exit.show()
            self.warning_text.display_text = self.warning_text.EXIT_TEXT
            self.post_message(EnterExchangeMode())
            self.scheduler.start()

        elif event.button.id == "button-exchange_mode-exit":
            self.button_refresh.enable()
            self.button_exit.hide()
            self.button_enter.show()
            self.warning_text.display_text = self.warning_text.ENTER_TEXT
            self.post_message(ExitExchangeMode())
            self.scheduler.shutdown()

        elif event.button.id == "button-exchange_mode-refresh":
            await self.update_data()

    async def _on_mount(self, event: events.Mount) -> None:
        self.scheduler.add_listener(self.on_executed, EVENT_JOB_EXECUTED)
        await self.update_data()


class ExchangeResultRow(UnClickableItem):
    """
    兑换结果行
    """
    DEFAULT_CSS = """
    ExchangeResultRow {
        border: round #666;
        padding: 1;
        height: auto;
        width: 1fr;
        layout: horizontal;
    }
    
    ExchangeResultRow Container {
        width: 1fr;
        height: auto;
        border: round #666;
        padding: 1;
        width: 1fr;
    }
    """
    rows: Dict[ExchangePlan, "ExchangeResultRow"] = {}
    """所有的兑换结果行"""

    def __init__(self, plan: ExchangePlan):
        """
        :param plan: 兑换计划
        """
        super().__init__()
        self.plan = plan
        """兑换计划"""
        self.result_preview = Container()
        """兑换结果字样预览"""
        self.rows.setdefault(plan, self)

    @classmethod
    def get_result_static(cls, text: str):
        """
        获取一个带有边框的Static 用于显示兑换结果
        """
        static = Static(text)
        static.styles.border = "round", "#666"
        static.styles.width = "1fr"
        return static

    def compose(self) -> ComposeResult:
        static = Static(f"[list]"
                        f"\n👓 米游社账号 - [bold green]{self.plan.account.bbs_uid}[/]"
                        f"\n📦 商品名称 - [bold green]{self.plan.good.goods_name}[/]"
                        f"\n📅 兑换时间 - [bold green]{self.plan.good.time_text}[/]"
                        f"\n🎮 游戏UID - [bold green]{self.plan.game_record.game_role_id if self.plan.game_record is not None else '[yellow]无需设置[/]'}[/]"
                        f"\n📮 收货地址 - [bold green]{self.plan.address.addr_ext if self.plan.address is not None else '[yellow]无需设置[/]'}[/]"
                        f"\n[/list]")
        static.styles.width = "2fr"
        yield static
        yield self.result_preview


class ExchangeModePing(Static):
    """
    兑换模式 Ping 结果的文本
    """
    DEFAULT_CSS = """
    ExchangeModePing {
        margin: 1 0;
    }
    """
    DEFAULT_VALUE = False
    ping_value: reactive[Union[float, bool, None]] = reactive(DEFAULT_VALUE)

    def render(self) -> RenderableType:
        return f"⚡ Ping | 商品兑换API服务器 [yellow]{_get_api_host() or 'N/A'}[/]" \
               f" - 延迟 [bold green]{round(self.ping_value, 2) or 'N/A'}[/] ms"

    def update_ping(self, event: JobExecutionEvent):
        """
        更新 Ping 值
        """
        if event.job_id == "exchange-connection_test":
            self.ping_value = event.retval

    def _on_mount(self, event: events.Mount) -> None:
        ExchangeModeView.scheduler.add_listener(self.update_ping, EVENT_JOB_EXECUTED)
